import os
import tempfile

import numpy as np
import xarray as xr

from typhon.files import NetCDF4


class TestNetCDF4:
    def test_dimension_mapping(self):
        """
        If a subgroup has not defined a dimension, but its parent group has one
        with the same size and name, the subgroup should use that one.
        Otherwise it should use the one of the subgroup (test with dim1 in the
        root and subgroups).
        """
        fh = NetCDF4()

        with tempfile.TemporaryDirectory() as tdir:
            tfile = os.path.join(tdir, 'testfile')
            before = xr.Dataset({
                "var1": ("dim1", np.arange(5)),
                "group1/var1": ("group1/dim1", np.arange(5)),
                "group1/var2": ("group1/dim2", np.arange(5)),
                "group1/subgroup1/var1":
                    ("group1/subgroup1/dim1", np.arange(5)),
                "group1/subgroup1/var2":
                    ("group1/subgroup1/dim2", np.arange(5)),
                "group2/var1": ("group2/dim1", np.arange(5)),
                "group2/subgroup1/var1":
                    ("group1/subgroup1/dim1", np.arange(5)),
                "group3/var1": ("group3/dim1", np.arange(10)),
            }, coords={
                "dim1": ("dim1", np.arange(5)),
                "group1/dim1": ("group1/dim1", np.arange(5))
            })

            # Save the dataset and load it again:
            fh.write(before, tfile)
            after = fh.read(tfile)

            # How it should be after loading:
            check = xr.Dataset({
                "var1": ("dim1", np.arange(5)),
                "group1/var1": ("group1/dim1", np.arange(5)),
                "group1/var2": ("group1/dim2", np.arange(5)),
                "group1/subgroup1/var1": ("group1/dim1", np.arange(5)),
                "group1/subgroup1/var2": ("group1/dim2", np.arange(5)),
                "group2/var1": ("dim1", np.arange(5)),
                "group2/subgroup1/var1": ("dim1", np.arange(5)),
                "group3/var1": ("group3/dim1", np.arange(10)),
            }, coords={
                "dim1": ("dim1", np.arange(5)),
                "group1/dim1": ("group1/dim1", np.arange(5))
            })

        assert after.equals(check)

    def test_scalar_masked(self):
        """Test if scalar masked values read OK

        Test for issue #277
        """

        fh = NetCDF4()

        with tempfile.TemporaryDirectory() as tdir:
            tfile = os.path.join(tdir, "testfile.nc")
            before = xr.Dataset({"a": xr.DataArray(42)})
            before["a"].encoding = {"_FillValue": 42}
            fh.write(before, tfile)
            after = fh.read(tfile)
            assert np.isnan(after["a"]) # fill value should become nan

    def test_times(self):
        """Test if times are read correctly
        """

        fh = NetCDF4()

        with tempfile.TemporaryDirectory() as tdir:
            tfile = os.path.join(tdir, "testfile.nc")
            before = xr.Dataset(
                    {"a":
                        xr.DataArray(
                            np.array(
                                ["2019-02-14T09:00:00", "2019-02-14T09:00:01"],
                                dtype="M8[ns]"))})
            before["a"].encoding = {
                    "units": "seconds since 2019-02-14 09:00:00",
                    "scale_factor": 0.1}
            fh.write(before, tfile)
            after = fh.read(tfile)
            assert np.array_equal(before["a"], after["a"])

    def test_scalefactor(self):
        """Test if scale factors written/read correctly
        """

        fh = NetCDF4()

        with tempfile.TemporaryDirectory() as tdir:
            tfile = os.path.join(tdir, "testfile.nc")
            before = xr.Dataset(
                    {"a":
                        xr.DataArray(
                            np.array([0.1, 0.2]))})
            before["a"].encoding = {
                    "scale_factor": 0.1,
                    "_FillValue": 42,
                    "dtype": "int16"}
            fh.write(before, tfile)
            after = fh.read(tfile)
            assert np.allclose(before["a"], after["a"])



# -*- coding: utf-8 -*-
"""Testing the functions in typhon.atmosphere.
"""
import numpy as np
import pytest

from typhon.physics import atmosphere


class TestAtmosphere:
    """Testing the atmosphere functions."""
    def test_integrate_water_vapor_hse(self):
        """Test the IWV calculation in hydrostatic equilibrium."""
        p = np.linspace(1000e2, 500e2, 10)
        vmr = np.linspace(0.025, 0.0025, p.size)

        iwv = atmosphere.integrate_water_vapor(vmr, p)

        assert np.allclose(iwv, 43.8845)

    def test_integrate_water_vapor_nonhse(self):
        """Test the IWV calculation in non-hydrostatic atmosphere."""
        p = np.linspace(1000e2, 500e2, 10)
        T = 288 * np.ones(p.size)
        z = np.linspace(0, 5000, p.size)
        vmr = np.linspace(0.025, 0.0025, p.size)

        iwv = atmosphere.integrate_water_vapor(vmr, p, T, z)

        assert np.allclose(iwv, 42.4062)

    def test_integrate_water_vapor_ivalidin(self):
        """Test invalid number of arguments to IWV calculation."""
        dummy = np.ones(10)

        with pytest.raises(ValueError):
            atmosphere.integrate_water_vapor(dummy, dummy, dummy)

    def test_integrate_water_vapor_multi(self):
        """Test multidimensional IWV calculation."""
        p = np.linspace(1000e2, 500e2, 10)
        vmr = np.linspace(0.025, 0.0025, p.size)
        vmr_multi = np.repeat(vmr[np.newaxis, :], 5, axis=0)

        iwv = atmosphere.integrate_water_vapor(vmr_multi, p, axis=1)

        assert np.allclose(iwv, np.repeat(43.8845, 5))

    def test_vmr2relative_humidity(self):
        """Test conversion from VMR into relative humidity."""
        rh = atmosphere.vmr2relative_humidity(0.025, 1013e2, 300)

        assert np.allclose(rh, 0.7160499)

    def test_relative_humidity2vmr(self):
        """Test conversion from relative humidity into VMR."""
        vmr = atmosphere.relative_humidity2vmr(0.75, 1013e2, 300)

        assert np.allclose(vmr, 0.0261853)

    def test_vmr_rh_conversion(self):
        """Check the consistency of VMR and relative humidity conversion.

        Converting VMR into relative humidity and back to ensure that both
        functions yield consistent results.
        """
        rh = atmosphere.vmr2relative_humidity(0.025, 1013e2, 300)
        vmr = atmosphere.relative_humidity2vmr(rh, 1013e2, 300)

        assert np.allclose(vmr, 0.025)

    def test_moist_lapse_rate(self):
        """Test calculation of moist-adiabatic lapse rate."""
        gamma = atmosphere.moist_lapse_rate(1000e2, 300)

        assert np.isclose(gamma, 0.00367349)

    def test_standard_atmosphere(self):
        """Test International Standard Atmosphere."""
        isa = atmosphere.standard_atmosphere

        assert np.isclose(isa(0), 288.1831)  # Surface temperature
        assert np.isclose(isa(81e3), 194.5951)  # Test extrapolation
        # Test call with ndarray.
        assert np.allclose(isa(np.array([0, 15e3])),
                           np.array([288.1831, 216.65]))

    def test_standard_atmosphere_pressure(self):
        """Test International Standard Atmosphere."""
        isa = atmosphere.standard_atmosphere

        assert np.isclose(isa(1000e2, coordinates='pressure'), 288.0527)

    def test_pressure2height(self):
        """Test conversion from atmospheric pressure to height."""
        p = np.array([1000e2, 750e2, 500e2, 100e2])

        z = atmosphere.pressure2height(p)
        z_ref = np.array([0., 2358.129, 5473.647, 15132.902])

        assert np.allclose(z, z_ref)

    def test_pressure2height_with_T(self):
        """Test conversion from atmospheric pressure to height."""
        p = np.array([1000e2, 750e2, 500e2, 100e2])
        T = np.array([288, 275, 255, 215])

        z = atmosphere.pressure2height(p, T=T)
        z_ref = np.array([0., 2360.809, 5482.749, 15135.793])

        assert np.allclose(z, z_ref)



"""
Jumping rules for MCMC
======================

This module contains classes that implement jumping rules for MCMC.

"""
import numpy as np

class RandomWalk:
    """
    Random walk jumping function for MCMC.
    """
    def __init__(self, covmat):
        """
        Create a random walker with given covariance.

        Args:
            covmat: A square, 2D numpy.array that should be used as covariance
            matrix for the random walk.
        """
        try:
            self.covmat = np.asarray(covmat)

            # covmat must be a matrix.
            if not self.covmat.ndim == 2:
                raise Exception("Covariance matrix must be 2 dimensional.")

            # covmat must be square.
            self.n = covmat.shape[0]
            if not self.n == covmat.shape[1]:
                raise Exception("Covariance matrix must be square.")

        except:
            raise Exception("covmat argument must be convertible to a numpy array.")

    def step(self, x):
        """
        Generates a new step x_new from the current position x.

        Args:
            x: The current position of the random walker.
        """
        return np.random.multivariate_normal(x.ravel(), self.covmat).reshape(x.shape)

    def update(self, hist):
        """
        Update covariance from a sequence of samples.

        This computes the covariance matrix of a given sequence of samples,
        scales it by  2.4/sqrt(n))^2 and sets it to the covariance matrix
        to be used for the random walk.

        Args:
            hist: 2D numpy array with shape (m,n) where m is number of steps in
            the sequence and n is the number of dimensions of the parameter
            space.
        """
        if not hist.shape[1] == self.n:
            raise Exception("Provided array does not have the expected dimensions.")
        mean = np.mean(hist, axis=0, keepdims=True)
        d = hist - mean
        s = np.dot(np.transpose(d), d) / hist.shape[0]
        self.covmat = (2.4 / np.sqrt(self.n)) ** 2 * s



"""The agendas submodule.

This module provides the Agenda class, which is used to represent parsed
controlfiles and can be executed on a given `Workspace` object.

"""

import ctypes as c
import numpy  as np
import os

from typhon.arts.workspace.api import find_controlfile, arts_api
from typhon.arts.workspace.output import CoutCapture

class Agenda:
    def __init__(self, ptr):
        """ Initialize Agenda object from pointer to C API Agenda object.
        Args:
            ptr(c.c_void_p): Pointer to Agenda object created with the parse_agenda
            method of the ARTS C API.
        """
        self.ptr       = ptr

    @classmethod
    def create(cls, name):
        """
        Create agenda with given name.

        Parameters:

            name(str): The name of the agenda to create.

        Returns:

            The newly created agenda object with the given name.

        """
        ptr = arts_api.create_agenda(name.encode())
        return Agenda(ptr)

    def clear(self):
        """ Reset agenda to be empty."""
        arts_api.agenda_clear(self.ptr)

    def append(self, agenda):
        """Append agenda to this agenda.

        Parameters:

            agenda(:class:`Agenda`): The agenda to append to this agenda.

        Raises:

            Exception if :code:`agenda` is not of the :class:`Agenda`
        """

        if not isinstance(agenda, Agenda):
            raise Exception("Agenda to append must be of type Agenda.")

        arts_api.agenda_append(self.ptr, agenda.ptr)

    def add_method(*args, **kwargs):
        """
        Add a workspace method call to the agenda.

        Parameters:

            ws(typhon.arts.workspace.Workspace): A (dummy) workspace object.

            m(typhon.arts.workspace.WorkspaceMethod): The method to add to the
            agenda

            *args:  Positional arguments of the WSM call to add to the agenda.

            **kwargs: Key-word arguments of the WSM call to add to the agenda.
        """
        from typhon.arts.workspace.variables import group_names

        if len(args) < 3:
            raise Exception("Need at least self, a workspace and the method to"
                            " add as arguments.")
        self = args[0]
        ws   = args[1]
        m    = args[2]
        m_id, args_out, args_in, temps = m._parse_output_input_lists(ws,
                                                                     args[3:],
                                                                     kwargs)
        arg_out_ptr = c.cast((c.c_long * len(args_out))(*args_out),
                             c.POINTER(c.c_long))
        arg_in_ptr = c.cast((c.c_long * len(args_in))(*args_in),
                            c.POINTER(c.c_long))
        if not m.name[-3:] == "Set" or not m.name[:-3] in group_names:

            for t in temps:
                arts_api.agenda_insert_set(ws.ptr, self.ptr, t.ws_id, t.group_id)

            arts_api.agenda_add_method(c.c_void_p(self.ptr), m_id,
                                       len(args_out), arg_out_ptr,
                                       len(args_in), arg_in_ptr)
        else:
            from typhon.arts.workspace.variables import WorkspaceVariable

            name_out = WorkspaceVariable.get_variable_name(args_out[0])
            name_in = WorkspaceVariable.get_variable_name(args_in[0])
            wsv_out = getattr(ws, name_out)
            wsv_in  = getattr(ws, name_in)

            ws.Copy(wsv_out, wsv_in)

            group_id = arts_api.get_variable(args_out[0]).group
            arts_api.agenda_insert_set(ws.ptr, self.ptr, args_out[0], group_id)

    def add_callback(self, f):
        """
        Add a Python callback to the agenda.

        The function f must except one argument, which is the the pointer to
        the workspace object on which the callback is executed.

        Parameters:

            f(callable): Python callable.

        """
        callback = c.CFUNCTYPE(None, c.c_void_p)(f)
        arts_api.agenda_insert_callback(self.ptr, callback)
        arts_api.callbacks += [callback]


    def execute(self, ws):
        """ Execute this agenda on the given workspace.
        Args:
            ws(Workspace): Workspace object on wich to execute the agenda.
        Raises:
            Exception: If execution of agenda on workspace fails.
        """
        with CoutCapture(ws):
            e = arts_api.execute_agenda(ws.ptr, self.ptr)
        if (e):
            raise Exception("Error during execution of Agenda:\n" + e.decode("utf8"))

    def __to_value_struct__(self):
        return {'ptr' : self.ptr}

    def __del__(self):
        """Destroys ARTS C API Agenda object associated with this Agenda object."""
        try:
            arts_api.destroy_agenda(self.ptr)
        except:
            pass

    @staticmethod
    def parse(name):
        """Parse controlfile and return agenda representing the agenda.

        Due to the way how ARTS works, agendas may not define WSVs with
        the same name. In this case, parsing of the agenda will fail.
        It is therefore not possible to parse an agenda twice which defines
        new workspace variables. In this case the user will have to keep track
        of the fist parsed agenda object.

        Args:
            name(str): Name of the control file. Is looked up recursively in the path
            specified by the ARTS_INCLUDE_PATH environmental variable.
        Raises:
            Exception: If parsing of the controlfile fails.
        """
        path = find_controlfile(name)
        ptr = arts_api.parse_agenda(path.encode())
        if not ptr:
            e = arts_api.get_error().decode("utf8")
            raise Exception("Error during parsing of controlfile " + str(path) + ":\n" + e)
        return Agenda(ptr)



# -*- coding: utf-8 -*-
"""Testing the basic mathematical functions.
"""
import numpy as np

import pytest

from typhon import math


class TestCommon:
    """Testing common mathematical functions."""
    def test_integrate_column(self):
        """Test numerical intergration of array elements.

        As no coordinates are given, the function is expected to integrate
        along the indices.
        """
        t = math.integrate_column(np.arange(5))

        assert(np.isclose(t, 8))

    def test_integrate_column_coordinates(self):
        """Test numerical intergration of array elements along coordinates."""
        x = np.linspace(0, 1, 5)
        t = math.integrate_column(np.arange(5), x)

        assert(np.isclose(t, 2))

    def test_interpolate_halflevels(self):
        ref = np.array([0.5, 1.5, 2.5])
        t = math.interpolate_halflevels(np.arange(4))

        assert(np.allclose(t, ref))

    def test_sum_digits(self):
        """Test calculating several digit sums."""
        assert(math.sum_digits(1) == 1)
        assert(math.sum_digits(10) == 1)
        assert(math.sum_digits(100) == 1)
        assert(math.sum_digits(123) == 6)
        assert(math.sum_digits(2**8) == 13)

    def test_nlogspace(self):
        """Test creation of logarithmic spaced array."""
        ref = np.array([1, 10, 100])
        t = math.nlogspace(1, 100, 3)

        assert(np.allclose(t, ref))

    def test_promote_maximally(self):
        """Test copying to high precision dtype."""
        x = np.array(1, dtype='int8')
        t = math.promote_maximally(x)

        assert(t.dtype == 'int64')

    def test_squeezable_logspace(self):
        """Test creation of squeezable logspace."""
        ref = np.array([1, 3.16227766, 100])
        t = math.squeezable_logspace(1, 100, 3, squeeze=0.5)

        assert(np.allclose(t, ref))

    def test_squeezable_logspace_nosqueeze(self):
        """Test creation of non-squeezed logspace.

        Without squeezing, results should be equal to normal ``nlogspace``.
        """
        ref = np.array([1, 10, 100])
        t = math.squeezable_logspace(1, 100, 3)

        assert(np.allclose(t, ref))

    def test_squeezable_logspace_fixpointbounds(self):
        """Test ValueError if fixpoint is out of bounds."""
        with pytest.raises(ValueError):
            math.squeezable_logspace(100, 1, fixpoint=1.1)

    def test_squeezable_logspace_squeezebounds(self):
        """Test ValueError if squeeze is out of bounds."""
        with pytest.raises(ValueError):
            math.squeezable_logspace(100, 1, squeeze=2.01)

class TestArray:
    """Testing array related mathematical functions."""
    def test_argclosest(self):
        x = np.arange(10)

        idx = math.array.argclosest(x, value=5.4)
        assert idx == 5

        idx, val = math.array.argclosest(x, value=5.4, retvalue=True)
        assert idx == 5 and val == 5.



# -*- coding: utf-8 -*-
from scipy.constants import h, c, k
import numpy as np


def PopuSource(LowerPop, UpperPop, LowerDeg, UpperDeg, Freq_c):
    sij = (2.*h*Freq_c**3)/(c**2*(LowerPop*UpperDeg/UpperPop/LowerDeg-1.))
    return sij


def PopuSource_AB(LowerPop, UpperPop, Aul, Bul, Blu):
    sij = (UpperPop*Aul)/(LowerPop*Blu - UpperPop*Bul)
    return sij


"""
PopuSource(Ite_pop[0][0][0],
           Ite_pop[0][0][1],
           MolPara.weighti[0],
           MolPara.weighti[1],
           MolPara.freqi[1][0]*1.e9)
PopuSource_AB(Ite_pop[0][0][0],
              Ite_pop[0][0][1],
              MolPara.ai[1,0],
              MolPara.bul[1,0],
              MolPara.blu[1,0])
MolPara.blu[1,0]
MolPara.bul[1,0]
MolPara.ai[1,0]
Ite_pop[0][0][0]
Ite_pop[0][0][1]
MolPara.weighti[1]
MolPara.weighti[0]
MolPara.freqi[1][0]*1.e9
"""


def Bv_T(Freq, T):
    # brbr = 1  # .e7*1.e-4
    Bv_out = 2.*h*Freq**3/c**2/(np.exp(h*Freq/k/T)-1.)
    return Bv_out




# -*- coding: utf-8 -*-

"""Miscellaneous convenience functions.
"""

import ast
import functools
import operator
import os
import shutil
import subprocess
import collections
import itertools
from numbers import Number
import traceback

from warnings import warn
from functools import (partial, wraps)

import xarray
import pandas
import numpy as np


__all__ = [
    'deprecated',
    'extract_block_diag',
    'safe_eval',
    'unique',
    'path_append',
    'path_prepend',
    'path_remove',
    'get_time_dimensions',
    'get_time_coordinates',
    'concat_each_time_coordinate',
    'undo_xarray_floatification',
    'image2mpeg',
    'stack_xarray_repdim',
    'split_units',
    'reraise_with_stack',
    'get_xarray_groups',
    'get_xarray_group',
    'add_xarray_groups',
    'to_array',
]


def deprecated(func=None, new_name=None, message=None):
    """Decorator which can be used to mark functions as deprecated.

    Examples:
        Calling ``foo()`` will raise a ``DeprecationWarning``.

        >>> @deprecated
        ... def deprecated_function():
        ...     pass

        Display message with additional information:

        >>> @deprecated(message='Additional information message.')
        ... def deprecated_function():
        ...     pass
    """
    # Return partial when no arguments are passed.
    # This allows a plain call of the decorator.
    if func is None:
        return partial(deprecated, new_name=new_name, message=message)

    # Build warning message (with optional information).
    msg = f'\nCall to deprecated function `{func.__name__}`.'

    if new_name is not None:
        msg += f' Use `{new_name}` instead.'

    if message is not None:
        msg += f'\n{message}'

    # Wrapper that prints the warning before calling the deprecated function.
    @wraps(func)
    def wrapper(*args, **kwargs):
        warn(msg, category=DeprecationWarning, stacklevel=2)
        return func(*args, **kwargs)

    if wrapper.__doc__ is None:
        wrapper.__doc__ = ''

    # Lines added to the docstring need to be indented with four spaces as this
    # is technically the case for all lines except the first one.
    # If this is not done, the Sphinx build will produce wrong results as the
    # relative indentation of the added lines and the original docstring does
    # not match.
    wrapper.__doc__ += (
        '\n\n    .. warning::\n       Function is deprecated'
        ' and will be removed in a future version.'
    )

    if new_name is not None:
        wrapper.__doc__ += f' Use :func:`{new_name}` instead.'

    if message is not None:
        wrapper.__doc__ += f'\n\n       {message}'

    return wrapper


def extract_block_diag(M, n):
    """Extract diagonal blocks from square Matrix.

    Args:
        M (np.array): Square matrix.
        n (int): Number of blocks to extract.

    Example:
        >>> foo = np.array([[ 1.,  1.,  0.,  0.],
        ... [ 1.,  1.,  0.,  0.],
        ... [ 0.,  0.,  2.,  2.],
        ... [ 0.,  0.,  2.,  2.]])
        >>> extract_block_diag(foo, 2)
        [array([[ 1.,  1.],
                [ 1.,  1.]]), array([[ 2.,  2.],
                [ 2.,  2.]])]

    """
    return [np.split(m, n, axis=1)[i] for i, m in enumerate(np.split(M, n))]

# This code, or a previous version thereof, was posted by user 'J. F.
# Sebastian' on http://stackoverflow.com/a/9558001/974555 on 2012-03-04
# and is dual-licensed under CC-BY-SA 3.0 and MIT, as confirmed at
# https://stackoverflow.com/questions/2371436/evaluating-a-mathematical-expression-in-a-string/9558001?noredirect=1#comment76927447_9558001
# on 2017-07-07

operators = {ast.Add: operator.add,
             ast.Sub: operator.sub,
             ast.Mult: operator.mul,
             ast.Div: operator.truediv,
             ast.Pow: operator.pow,
             ast.BitXor: operator.xor,
             ast.USub: operator.neg}


def safe_eval(expr):
    """Safely evaluate string that may contain basic arithmetic
    """

    return _safe_eval_node(ast.parse(expr, mode="eval").body)


def _safe_eval_node(node):
    if isinstance(node, ast.Num):  # <number>
        return node.n
    elif isinstance(node, ast.BinOp):  # <left> <operator> <right>
        return operators[type(node.op)](
            _safe_eval_node(node.left), _safe_eval_node(node.right))
    elif isinstance(node, ast.UnaryOp):  # <operator> <operand> e.g., -1
        return operators[type(node.op)](_safe_eval_node(node.operand))
    else:
        raise TypeError(node)
# End of snippet derived from http://stackoverflow.com/a/9558001/974555


def unique(seq):
    """Remove duplicates from list whilst keeping the original order

    Notes:
        If you do not care about keeping the order, use this code:
        >>>> list(set([0, 5, 1, 2, 0, 3, 1,]))
        [0, 1, 2, 3, 5]

    This code is taken from https://stackoverflow.com/a/480227.

    Args:
        seq: A sequence (list, etc.) of elements.

    Returns:
        A list with unique items with original order.

    Examples:
        >>>> unique([0, 5, 1, 2, 0, 3, 1,])
    """
    seen = set()
    seen_add = seen.add
    return [x for x in seq if not (x in seen or seen_add(x))]


def path_append(dirname, path='PATH'):
    """Append a directory to environment path variable.

    Append entries to colon-separated variables (e.g. the system path).
    If the entry is already in the list, it is moved to the end.
    A path variable is set, if not existing at function call.

    Parameters:
        dirname (str): Directory to add to the path.
        path (str): Name of the path variable to append to.
            Defaults to the system path 'PATH'.
    """
    if path in os.environ:
        dir_list = os.environ[path].split(os.pathsep)
        if dirname in dir_list:
            dir_list.remove(dirname)
        dir_list.append(dirname)
        os.environ[path] = os.pathsep.join(dir_list)
    else:
        os.environ[path] = dirname


def path_prepend(dirname, path='PATH'):
    """Prepend a directory to environment path variable.

    Append entries to colon-separated variables (e.g. the system path).
    If the entry is already in the list, it is moved to the end.
    A path variable is set, if not existing at function call.

    Parameters:
        dirname (str): Directory to add to the path.
        path (str): Name of the path variable to append to.
            Defaults to the system path 'PATH'.
    """
    if path in os.environ:
        dir_list = os.environ[path].split(os.pathsep)
        if dirname in dir_list:
            dir_list.remove(dirname)
        dir_list.insert(0, dirname)
        os.environ[path] = os.pathsep.join(dir_list)
    else:
        os.environ[path] = dirname


def path_remove(dirname, path='PATH'):
    """Remove a directory from environment path variable.

    Remove entries from colon-separated variables (e.g. the system path).
    If the path variable is not set, nothing is done.

    Parameters:
        dirname (str): Directory to add to the path.
        path (str): Name of the path variable to append to.
            Defaults to the system path 'PATH'.
    """
    if path in os.environ:
        dir_list = os.environ[path].split(os.pathsep)
        dir_list.remove(dirname)
        os.environ[path] = os.pathsep.join(dir_list)


def get_time_dimensions(ds):
    """From a xarray dataset or dataarray, get dimensions corresponding to time coordinates

    """

    return {k for (k, v) in ds.coords.items() if k in ds.dims and v.dtype.kind == "M"}


def get_time_coordinates(ds):
    """From a xarray dataset or dataarray, get coordinates with at least 1 time dimension

    """

    time_dims = get_time_dimensions(ds)
    return {k for (k, v) in ds.coords.items() if set(v.dims)&time_dims}


# Any commits made to this module between 2015-05-01 and 2017-03-01
# by Gerrit Holl are developed for the EC project “Fidelity and
# Uncertainty in Climate Data Records from Earth Observations (FIDUCEO)”.
# Grant agreement: 638822.  This specifically applies to the function
# concat_each_time_coordinate.
#
# All those contributions are dual-licensed under the MIT license for use
# in typhon, and the GNU General Public License version 3.

def concat_each_time_coordinate(*datasets):
    """Concatenate xarray datasets along each time coordinate

    Given two or more xarray datasets, concatenate seperately data
    variables with different time coordinates.  For example, one might
    have dimensions 'scanline' and 'calibration_cycle' that are each along
    time coordinates.  Data variables may have dimension either scanline
    or calibration_cycle or neither, but not both.  Both correspond to
    coordinates a datetime index.  Ordinary xarray.concat along either
    dimension will broadcast the other one in a way similar to repmat,
    thus exploding memory usage (test case for one FCDR HIRS granule: 89
    MB to 81 GB).  Instead, here, for each data variable, we will
    concatenate only along at most one time coordinate.

    Arguments:

        *datasets: xarray.Dataset objects to be concatenated
    """

    time_coords = get_time_coordinates(datasets[0])
    time_dims = get_time_dimensions(datasets[0])
    # ensure each data-variable has zero or one of those time coordinates
    # as dimensions
    for ds in datasets:
        if not all([len(set(v.dims) & time_coords) <= 1
                    for (k, v) in ds.data_vars.items()]):
            raise ValueError("Found vars with multiple time coords")

    new_sizes = {k: sum(g.dims[k] for g in datasets)
                     if k in time_coords
                     else datasets[0].dims[k]
                 for k in datasets[0].dims.keys()}
    # note data vars per time coordinate
    time_vars = {k: (set(v.dims)&time_coords).pop() for (k, v) in datasets[0].variables.items() if set(v.dims)&time_coords}
    time_vars_per_time_dim = {k: {vn for (vn, dn) in time_vars.items() if dn==k} for k in time_coords}
    untimed_vars = datasets[0].data_vars.keys() - time_vars.keys()

    # allocate new
    new = xarray.Dataset(
        {k: (v.dims,
             np.zeros(shape=[new_sizes[d] for d in v.dims],
                         dtype=v.dtype))
                for (k, v) in datasets[0].data_vars.items()})
    # coordinates cannot be set in the same way so need to be allocated
    # separately
    new_coords = {k: xarray.DataArray(
                        np.zeros(shape=[new_sizes[d] for d in v.dims],
                                 dtype=datasets[0][k].dtype),
                        dims=v.dims)
                    for (k, v) in datasets[0].coords.items()}

    # copy over untimed vars
    for v in untimed_vars:
        new[v].values[...] = datasets[0][v].values

    # and untimed coords
    for c in datasets[0].coords.keys() - time_coords:
        new_coords[c][...] = datasets[0].coords[c]

    # keep track of progress per time dimension
    n_per_dim = dict.fromkeys(time_coords, 0)
    # copy timed vars dataset by dataset
    for ds in datasets:
        for (v, timedim) in time_vars.items():
            ncur = n_per_dim[timedim]
            nnew_cur = ds.dims[timedim]
            if nnew_cur == 0:
                # nothing to fill, but prevent
                # https://github.com/pydata/xarray/issues/1329
                continue
            slc = {dim: slice(ncur, ncur+nnew_cur)
                        if dim==timedim else slice(None)
                   for dim in ds[v].dims}
            if v in time_coords: # time coordinate
                new_coords[v][slc] = ds[v]
            else:
                new[v].loc[slc] = ds[v]
        for timedim in time_dims:
            n_per_dim[timedim] += ds.dims[timedim]
    # copy attributes
    new.attrs.update(**datasets[0].attrs)
    for k in new.variables.keys():
        new[k].attrs.update(**datasets[0][k].attrs)
        new[k].encoding.update(**datasets[0][k].encoding)
    return new.assign_coords(**new_coords)

def undo_xarray_floatification(ds, fields=None):
    """convert floats back to ints in xarray dataset where appropriate

    When xarray opens a NetCDF file with the default decode_cf=True,
    any integer values that have a _FillValue set are converted to float,
    such that any _FillValue-set values can be set to nan.  Some datasets
    may have such _FillValue set even though they are never used.
    In this case, it may be desirable to convert those values back to
    the original dtype (which is preserved in the .encoding attribute),
    for example, when those integers are intended to be used as indices.
    This function takes an xarray Dataset, checks all the variables which
    originally have an integer dtype and a fillvalue set, and converts
    those back to int.  Optionally only a subset of those is converted.

    Use this function only when those fill values are not used.  Behaviour
    when fill values are actually used is undefined.

    Parameters:

        ds (xarray.Dataset): xarray dataset to be converted.  Will be
        copied.

        fields (Collection or None): Describes what fields shall be
            converted.  If not given or None (default), convert all fields
            that were originally ints but converted to float due to having a
            _FillValue set.  Even when given, only fields meeting those
            criteria will be converted.

    Returns:
        The same dataset but with changes as described above.
    """

    to_correct = {k for (k, v) in ds.data_vars.items()
        if v.encoding.get("dtype", np.dtype("O")).kind[0] in "ui" and
        not v.dtype.kind in "uiMm"} # don't convert datetime/deltas

    if fields is not None:
        to_correct &= fields

    ds2 = ds.copy()
    for k in to_correct:
        ds2[k] = ds[k].astype(ds[k].encoding["dtype"])
        ds2[k].encoding.update(ds[k].encoding)

    return ds2

def image2mpeg(glob, outfile, framerate=12, resolution='1920x1080'):
    """Combine image files to a video using ``ffmpeg``.

    Notes:
        The function is tested for ``ffmpeg`` versions 2.8.6 and 3.2.2.

    Parameters:
        glob (str): Glob pattern for input files.
        outfile (str): Path to output file.
            The file fileformat is determined by the extension.
        framerate (int or str): Number of frames per second.
        resolution (str or tuple): Video resolution given in width and height
            (``"WxH"`` or ``(W, H)``).

    Raises:
        Exception: The function raises an exception if the
            underlying ``ffmpeg`` process returns a non-zero exit code.

    Example:
        >>> image2mpeg('foo_*.png', 'foo.mp4')
    """
    if not shutil.which('ffmpeg'):
        raise Exception('``ffmpeg`` not found.')

    # If video resolution is given as tuple, convert it into string format
    # to directly pass it to ffmpeg later.
    if isinstance(resolution, tuple) and len(resolution) == 2:
        resolution = '{width}x{height}'.format(width=resolution[0],
                                               height=resolution[1])

    p = subprocess.run(
        ['ffmpeg',
         '-framerate', str(framerate),
         '-pattern_type', 'glob', '-i', glob,
         '-s:v', resolution,
         '-c:v', 'libx264',
         '-profile:v', 'high',
         '-crf', '20',
         '-pix_fmt', 'yuv420p',
         '-y', outfile
         ],
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
        universal_newlines=True
        )

    # If the subprocess fails, raise exception including error message.
    if p.returncode != 0:
        raise Exception(p.stderr)

def stack_xarray_repdim(da, **dims):
    """Like xarrays stack, but with partial support for repeated dimensions

    The xarray.DataArray.stack method fails when any dimension occurs
    multiple times, as repeated dimensions are not currently very well
    supported in xarray (2018-03-26).  This method provides a workaround
    so that stack can be used for an array where some dimensions are
    repeated, as long as the repeated dimensions are themselves not
    stacked.


    Parameters:
        da (DataArray): DataArray to operate on.
        **dims: Dimensions to stack.  As for xarray.DataArray.stack.
    """

    # make view of da without repeated dimensions
    cnt = collections.Counter(da.dims)
    D = {k: itertools.count() for k in cnt.keys()}
    tmpdims = []
    dimmap = {}
    for dim in da.dims:
        if cnt[dim] == 1:
            tmpdims.append(dim)
        else:
            newdim = "{:s}{:d}".format(dim, next(D[dim]))
            tmpdims.append(newdim)
            dimmap[newdim] = dim
    da2 = xarray.DataArray(da.values, dims=tmpdims)
    da2_stacked = da2.stack(**dims)
    # put back repeated dimensions with new coordinates
    da3 = xarray.DataArray(da2_stacked.values,
        dims=[dimmap.get(d, d) for d in da2_stacked.dims])
    da3 = da3.assign_coords(
        **{k: pandas.MultiIndex.from_product(
                    [da.coords[kk] for kk in dims[k]], names=dims[k])
                if k in dims else da.coords[k] for k in np.unique(da3.dims)})
    return da3


def split_units(value):
    """Splits a string into float number and potential unit

    References
        Taken from https://stackoverflow.com/a/30087094

    Args:
        value: String with number and unit.

    Returns
        A tuple of a float and unit string.

    Examples:

        >>> split_units("2GB")
        (2.0, 'GB')
        >>> split_units("17 ft")
        (17.0, 'ft')
        >>> split_units("   3.4e-27 frobnitzem ")
        (3.4e-27, 'frobnitzem')
        >>> split_units("9001")
        (9001.0, '')
        >>> split_units("spam sandwhiches")
        (0, 'spam sandwhiches')
        >>> split_units("")
        (0, '')
    """
    units = ""
    number = 0
    while value:
        try:
            number = float(value)
            break
        except ValueError:
            units = value[-1:] + units
            value = value[:-1]
    return number, units.strip()


def reraise_with_stack(func):
    """Make functions include the whole stack in raised exceptions

    Notes:
        This is a decorator function.

    When using the concurrent.futures module, the original traceback message
    gets lost, which makes it difficult to debug. This decorator solves the
    problem.

    References:
        Taken from https://stackoverflow.com/a/29357032.
    """

    @functools.wraps(func)
    def wrapped(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except Exception as e:
            traceback_str = traceback.format_exc()
            raise Exception(
                "Error occurred. Original traceback is\n%s\n" % traceback_str
            )

    return wrapped


def get_xarray_groups(dataset, only_names=False):
    """Get pseudo groups from xarray dataset object (only direct under root)

    xarray.Dataset objects does not allow the use of groups, but you can
    emulate them by using */* in the variable's name.

    Args:
        dataset: A xarray.Dataset object
        only_names: Return only the names of the groups.

    Returns:
        A set of group names if `only_names` is True. Otherwise, a dictionary
        with the group names and all its variables as xarray.Datasets objects.
    """

    groups = {
        var_name.split("/", 1)[0]
        for var_name in dataset.variables
        if "/" in var_name
    }

    if only_names:
        return groups

    return {
        group: get_xarray_group(dataset, group)
        for group in groups
    }


def get_xarray_group(dataset, group):
    """Get pseudo group from xarray.Dataset

    Args:
        dataset: A xarray.Dataset object with pseudo groups.
        group: The name of the group (can also be a subgroup).

    Returns:
        A xarray.Dataset with the pseudo group.
    """

    if not group.endswith("/"):
        group += "/"

    group_vars = [
        var
        for var in dataset.variables
        if var.startswith(group)
    ]

    if not group_vars:
        raise KeyError(f"The group {group} was not found!")

    return dataset[group_vars]


def add_xarray_groups(ds, **kwargs):
    """Add a xarray.Dataset as a subgroup to another xarray.Dataset

    Args:
        ds: The root xarray.Dataset.
        **kwargs: Keyword arguments: the key is the name of the group and the
            value must be a xarray.Dataset.

    Returns:
        `ds` with the added subgroups
    """
    datasets = [ds]

    for group_name, group in kwargs.items():
        group = group.rename(
            {
                var_name: "/".join([group_name, var_name])
                for var_name in group.variables
            },
        )

        # Add the group name also to the dimensions:
        group = group.rename({
            dim: "/".join([group_name, dim])
            for dim in group.dims
            if dim not in group.coords
        })

        datasets.append(group)

    return xarray.merge(datasets)


def to_array(item):
    """Convert item to numpy array

    Args:
        item: Can be a number, list, tuple or numpy array.

    Returns:
        `item` converted to a numpy array.
    """
    if isinstance(item, np.ndarray):
        return item

    if isinstance(item, Number):
        return np.array([item])
    else:
        return np.array(item)



import numpy as np

def trapz_inte_edge(y, x):
    """ Trapezoidal integration including edge grids

    Parameters:
        y: Array of y-axis value
        x: Array of x-axis value

        Returns:
        Area corresponded to each y (or x) value
            For Example,
             Area corresponded to at y_n is
            ..math:
                0.5*y_n ((x_{n} - x_{n-1}) + (x_{n+1} - x_{n}))
             Area corresponded to at y_0 (start point(edge)) is
            ..math:
                0.5*y_0(x_{1} - x_{0})
    """
    weight_x_0 = 0.5 * (x[1] - x[0])
    weight_x_f = 0.5 * (x[-1] - x[-2])
    weight_x_n = 0.5 * (x[1:-1] - x[:-2]) + 0.5 * (x[2:] - x[1:-1])
    weight_x = np.r_[weight_x_0, weight_x_n, weight_x_f]
    return weight_x*y



"""
The mcmc submodule.
===================

Contains the MCMC which implements the Marcov Chain Monte Carlo method to
sample from a posterior distribution of a retrieval problem.

References
==========

[1] Andrew Gelman et al., Bayesian Data Analysis, 3rd Edition

"""
import logging

import numpy as np


logger = logging.getLogger(__name__)


def r_factor(stats):
    """
    This computes the R-factor as defined in 'Bayesian Data Analysis'
    (Getlman et al.), Chapter 11. If the simulations have converged,
    the result of `r_factor` should be close to one.

    Args:
        stats: A list of arrays of statistics (scalar summaries) computed from
            serveral MCMC runs.
    """
    n = stats[0].size
    m = len(stats)

    vars  = np.array([np.var(s) for s in stats])
    means = np.array([np.mean(s) for s in stats])
    mean  = np.mean(means)

    b = n * np.var(means)
    w = np.mean(vars)

    var_p = (n - 1) / n * w + b / n
    return np.sqrt(var_p / w)

def variogram(stats, t):
    """
    Helper function that computes the variogram for a given lag t. The
    variogram is the mean of the mean squared sum of deviations of lag t of
    each sequence.

    Args:
        stats: A list of sequences
    """
    m = len(stats)
    n = stats[0].size
    return sum([np.sum((s[t+1:] - s[:n-t-1])**2) for s in stats]) / (m * (n - t))

def split(stats):
    """
    Splits a list of sequences in halves.

    Sequences generated from MCMC runs should be split in half in order to be
    able to properly diagnose mixing.

    Args:
        stats: A list of sequences
    """
    n = stats[0].size
    return [s[i * (n // 2) : (i + 1) * (n // 2)]
            for i in range(2) for s in stats]

def autocorrelation(stats):
    """
    Estimates the autocorrelation of a list of sequences from a MCMC run.
    This uses formula (11.7) in [1] to approximate the autocorrelation function
    for lags [0, n // 2].
    """
    n = stats[0].size

    vars  = np.array([np.var(s) for s in stats])
    means = np.array([np.mean(s) for s in stats])
    b = n * np.var(means)
    w = np.mean(vars)
    var_p = (n - 1) / n * w + b / n

    rho = np.zeros(n // 2)
    for t in range(n // 2):
        vt     = variogram(stats, t)
        rho[t] = 1.0 - 0.5 * vt / var_p
    return rho

def effective_sample_size(stats):
    """
    This estimates the effective sample size of independent samples from the
    posterior distribution using formula (11.8) in [1].
    """
    n = stats[0].size
    m = len(stats)

    vars  = np.array([np.var(s) for s in stats])
    means = np.array([np.mean(s) for s in stats])
    b = n * np.var(means)
    w = np.mean(vars)
    var_p = (n - 1) / n * w + b / n

    rho = np.zeros(n // 2)
    for t in range(n // 2):
        vt     = variogram(stats, t)
        rho[t] = 1.0 - 0.5 * vt / var_p
        if t > 2 and (rho[t-1] + rho[t-2]) < 0.0:
            break

    return m * n / (1.0 + 2.0 * sum(rho[:t-2]))

class MCMC:
    """
    The MCMC class represents an ongoing MCMC simulation. An MCMC object can be
    used to run a given number of MC steps, test the results for convergence
    and perform further calculations if necessary.
    """
    @staticmethod
    def _check_input(vars, py, stats):
        """
        Helper method that checks arguments provided to `__init__`.
        """
        if not type(vars) == list and len(vars) > 0:
            raise Exception("Argument vars must be of type list and have length"
                            + "> 0.")
        for v in vars:
            if not type(v) == list and len(v) == 3:
                raise Exception("Elements of argument vars must be tuples of "
                                + " mutable variables, jump functions and prior"
                                + " densities.")
            if not callable(v[1]):
                raise Exception("Non-callable object given as prior density.")
            if not callable(v[2]):
                raise Exception("Non-callable object given as jump function.")
        if not callable(py):
                raise Exception("Non-callable object given for conditional"
                                + "probability p(y | x).")
        for s in stats:
            if not callable(s):
                raise Exception("Non-callable object given as statistic.")


    def __init__(self, vars, y, ly, stats = []):
        """
        To construct an MCMC object, the user must provide a list of variables,
        prior distributions and likelihood functions, the measurement vector,
        a measurement likelihood and optionally a set of stats to evaluate at
        each step.

        Args:
           vars: A list of triples (v,l,j) containing a triple of a variable
               v, a prior likelihood function l so that `l(v)` yields a value
               proportional to the logarithm of the prior probability of value
               of v, and finally a jump function j, so that
               `v_new = j(ws, v_old)` yields a new value for the variable v
               and manipulates the :class:`~typhon.arts.workspace.Workspace`
               object ws so that a subsequent call to the yCalc WSM will compute
               the simulated measurement corresponding to the new value `v_new`
               of the variable v.
           y: The measured vector of brightness temperatures which must be
               consistent with the ARTS WSV y
           ly: The measurement likelihood such that `ly(y, yf)` gives
               the log of the probability that deviations between `y` and
               `yf` are due to measurement errors.
           stats: This is a list of statstics such that for each element
               s `s(ws)` is a scalar value computed on a given workspace.
        """
        MCMC._check_input(vars, ly, stats)
        self.y     = y
        self.vars  = vars
        self.ly    = ly
        self.stats = stats

    def eval_l(self, ws):
        """
        Evaluate the likelihood of the current state. This method
        simply computes and sums up the measurement likelihood and
        the prior likelihoods.

        Args:
           ws: A :class:`~typhon.arts.workspace.Workspace` object consistent
               with the current state of the MCMC run.
        """
        lxs = np.zeros(len(self.vars))
        ly  = self.ly(self.y, ws.y.value)
        for i, (x, l, _) in enumerate(self.vars):
            lxs[i] = l(x)
        return ly, lxs

    def step(self, ws, ly_old, lxs_old):
        """
        The performs a Gibbs step for a given variable. This will generate
        a candidate for the given variable using the corresponding jump
        function, call `yCalc()` on the
        :class:`~typhon.arts.workspace.Workspace` object `ws`

        Args:
           ws: A :class:`~typhon.arts.workspace.Workspace` object consistent
               with the current state of the MCMC run.
           ly_old: The measurement likelihood before the execution of
               new step
           lxs_old: The prior likelihoods for each of the variables
               that are being retrieved.
        """
        accepted = np.zeros((1, len(self.vars)), dtype=bool)
        lxs = np.zeros(lxs_old.shape)
        for i, ((x,l,j), lx_old) in enumerate(zip(self.vars, lxs_old)):

            # Generate new step
            x_new = j(ws, x)
            ws.yCalc()
            lx_new = l(x_new)
            ly_new = self.ly(self.y, ws.y.value)

            # Check Acceptance
            r = np.exp(lx_new + ly_new - lx_old - ly_old)
            if r > 1.0 or np.random.random() < r:
                x[:] = x_new
                accepted[0, i] = True
                ly_old = ly_new
                lxs[i] = lx_new
            else:
                j(ws, x, revert = True)
                accepted[0, i] = False
                lxs[i] = lxs_old[i]

        return accepted, ly_old, lxs

    def print_log(self, step, acceptance):
        """
        Prints log output to stdout.

        Args:
            step: The number of the current step
            acceptance: The array of bools tracking the acceptances for
                each simulation step.
        """
        if step > 0:
            ar = sum(acceptance / step)
        else:
            ar = 0.0

        logger.info("MCMC Step " + str(step) + ": " + "ar = " + str(ar))

    def warm_up(self, ws, x0s, n_steps):
        """
        Run a simulation of `n_steps` on a given workspace `ws`
        starting from start values `x0s`.

        Args:
            ws: A :class:`~typhon.arts.workspace.Workspace` object setup so that
                only a call to the `yCalc` WSM is necessary to perform a
                simulation
            x0s: A list of start values which is used to initialized the
                workspace by calling `j(x0)` for each `x0` in `x0s` and `j` is
                the jump function of the corresponding variable.
        """
        ls    = np.zeros(n_steps + 1)
        stats = np.zeros((n_steps + 1, len(self.stats)))
        hist  = [np.zeros((n_steps + 1,) + x.shape) for x,l,j in self.vars]
        acceptance = np.zeros((n_steps, len(self.vars)), dtype=bool)

        lxs = np.zeros(len(self.vars))
        ly  = 0.0

        # Set initial state.
        for i,((x, l, j), x0) in enumerate(zip(self.vars, x0s)):
            x[:] = j(ws, x0)
            hist[i][0,:] = x[:]

        # Evaluate likelihood
        ws.yCalc()
        ly, lxs = self.eval_l(ws)
        ls[0] = ly + sum(lxs)

        # Evaluate statistics
        for i,s in enumerate(self.stats):
            stats[0, i] = s(ws)

        for i1 in range(n_steps):
            acceptance[i1, :], ly, lxs = self.step(ws, ly, lxs)
            for i2, h in enumerate(hist):
                hist[i2][i1+1, :] = self.vars[i2][0][:]
            for i2,s in enumerate(self.stats):
                    stats[i1 + 1, i2] = s(ws)
            if (i1 % 10) == 0:
                self.print_log(i1, acceptance)
        self.ly_old    = ly
        self.lxs_old   = lxs
        self.stats_old = stats[-1,:]
        self.hist_old  = [h[-1,:] for h in hist]
        return hist, stats, ls, acceptance

    def run(self, ws, n_steps):
        """
        Run a simulation of `n_steps` on a given workspace `ws`
        starting from start values `x0s`.

        Args:
            ws: A :class:`~typhon.arts.workspace.Workspace` object setup so that
            only a call to the `yCalc` WSM is necessary to perform a simulation
            x0s: A list of start values which is used to initialized the
                workspace by calling `j(x0)` for each `x0` in `x0s` and `j` is
                the jump function of the corresponding variable.
        """
        ls    = np.zeros(n_steps)
        stats = np.zeros((n_steps, len(self.stats)))
        hist  = [np.zeros((n_steps,) + x.shape) for x,l,j in self.vars]
        acceptance = np.zeros((n_steps, len(self.vars)), dtype=bool)

        ly  = self.ly_old
        lxs = self.lxs_old

        for i1 in range(n_steps):
            acceptance[i1, :], ly, lxs = self.step(ws, ly, lxs)
            for i2, h in enumerate(hist):
                hist[i2][i1, :] = self.vars[i2][0][:]
            for i2,s in enumerate(self.stats):
                    stats[i1, i2] = s(ws)
            if (i1 % 10) == 0:
                self.print_log(i1, acceptance)
        self.ly_old  = ly
        self.lxs_old = lxs
        self.stats_old = stats[-1,:]
        self.hist_old  = [h[-1,:] for h in hist]
        return hist, stats, ls, acceptance



# -*- coding: utf-8 -*-

"""Functions directly related to atmospheric sciences.
"""
import numpy as np
from scipy.interpolate import interp1d

from typhon import constants
from typhon import math
from typhon.physics import thermodynamics


__all__ = [
    'relative_humidity2vmr',
    'vmr2relative_humidity',
    'integrate_water_vapor',
    'moist_lapse_rate',
    'standard_atmosphere',
    'pressure2height',
]


def relative_humidity2vmr(RH, p, T, e_eq=None):
    r"""Convert relative humidity into water vapor VMR.

    .. math::
        x = \frac{\mathrm{RH} \cdot e_s(T)}{p}

    Note:
        By default, the relative humidity is calculated with respect to
        saturation over liquid water in accordance to the WMO standard for
        radiosonde observations.
        You can use :func:`~typhon.physics.e_eq_mixed_mk` to calculate
        relative humidity with respect to saturation over the mixed-phase
        following the IFS model documentation.

    Parameters:
        RH (float or ndarray): Relative humidity.
        p (float or ndarray): Pressue [Pa].
        T (float or ndarray): Temperature [K].
        e_eq (callable): Function to calculate the equilibrium vapor
            pressure of water in Pa. The function must implement the
            signature ``e_eq = f(T)`` where ``T`` is temperature in Kelvin.
            If ``None`` the function :func:`~typhon.physics.e_eq_water_mk` is
            used.

    Returns:
        float or ndarray: Volume mixing ratio [unitless].

    See also:
        :func:`~typhon.physics.vmr2relative_humidity`
            Complement function (returns RH for given VMR).
        :func:`~typhon.physics.e_eq_water_mk`
            Used to calculate the equilibrium water vapor pressure.

    Examples:
        >>> relative_humidity2vmr(0.75, 101300, 300)
        0.026185323887350429
    """
    if e_eq is None:
        e_eq = thermodynamics.e_eq_water_mk

    return RH * e_eq(T) / p


def vmr2relative_humidity(vmr, p, T, e_eq=None):
    r"""Convert water vapor VMR into relative humidity.

    .. math::
        \mathrm{RH} = \frac{x \cdot p}{e_s(T)}

    Note:
        By default, the relative humidity is calculated with respect to
        saturation over liquid water in accordance to the WMO standard for
        radiosonde observations.
        You can use :func:`~typhon.physics.e_eq_mixed_mk` to calculate
        relative humidity with respect to saturation over the mixed-phase
        following the IFS model documentation.

    Parameters:
        vmr (float or ndarray): Volume mixing ratio,
        p (float or ndarray): Pressure [Pa].
        T (float or ndarray): Temperature [K].
        e_eq (callable): Function to calculate the equilibrium vapor
            pressure of water in Pa. The function must implement the
            signature ``e_eq = f(T)`` where ``T`` is temperature in Kelvin.
            If ``None`` the function :func:`~typhon.physics.e_eq_water_mk` is
            used.

    Returns:
        float or ndarray: Relative humidity [unitless].

    See also:
        :func:`~typhon.physics.relative_humidity2vmr`
            Complement function (returns VMR for given RH).
        :func:`~typhon.physics.e_eq_water_mk`
            Used to calculate the equilibrium water vapor pressure.

    Examples:
        >>> vmr2relative_humidity(0.025, 1013e2, 300)
        0.71604995533615401
    """
    if e_eq is None:
        e_eq = thermodynamics.e_eq_water_mk

    return vmr * p / e_eq(T)


def integrate_water_vapor(vmr, p, T=None, z=None, axis=0):
    r"""Calculate the integrated water vapor (IWV).

    The basic implementation of the function assumes the atmosphere
    to be in hydrostatic equilibrium.  The IWV is calculated as follows:

    .. math::
        \mathrm{IWV} = -\frac{1}{g} \int q(p)\,\mathrm{d}p

    For non-hydrostatic atmospheres, additional information
    on temperature and height are needed:

    .. math::
        \mathrm{IWV} = \int \rho_v(z)\,\mathrm{d}z

    Parameters:
        vmr (float or ndarray): Volume mixing ratio,
        p (float or ndarray): Pressue [Pa].
        T (float or ndarray): Temperature [K] (see ``z``).
        z (float or ndarray): Height [m]. For non-hydrostatic calculation
            both ``T`` and ``z`` have to be passed.
        axis (int): Axis to integrate along.

    Returns:
        float: Integrated water vapor [kg/m**2].
    """
    if T is None and z is None:
        # Calculate IWV assuming hydrostatic equilibrium.
        q = thermodynamics.vmr2specific_humidity(vmr)
        g = constants.earth_standard_gravity

        return -math.integrate_column(q, p, axis=axis) / g
    elif T is None or z is None:
        raise ValueError(
            'Pass both `T` and `z` for non-hydrostatic calculation of the IWV.'
        )
    else:
        # Integrate the water vapor mass density for non-hydrostatic cases.
        R_v = constants.gas_constant_water_vapor
        rho = thermodynamics.density(p, T, R=R_v)  # Water vapor density.

        return math.integrate_column(vmr * rho, z, axis=axis)


def moist_lapse_rate(p, T, e_eq=None):
    r"""Calculate the moist-adiabatic temperature lapse rate.

    Bohren and Albrecht (Equation 6.111, note the **sign change**):

    .. math::
        \frac{dT}{dz} =
            \frac{g}{c_p} \frac{1 + l_v w_s / RT}{1 + l_v^2 w_s/c_p R_v T^2}

    Parameters:
        p (float or ndarray): Pressure [Pa].
        T (float or ndarray): Temperature [K].
        e_eq (callable): Function to calculate the equilibrium vapor
            pressure of water in Pa. The function must implement the
            signature ``e_eq = f(T)`` where ``T`` is temperature in Kelvin.
            If ``None`` the function :func:`~typhon.physics.e_eq_water_mk` is
            used.

    Returns:
        float or ndarray: Moist-adiabatic lapse rate [K/m].

    Examples:
        >>> moist_lapse_rate(1013.25e2, 288.15)
        0.004728194612232855

    References:
        Bohren C. and Albrecht B., Atmospheric Thermodynamics, p. 287-92
    """
    if e_eq is None:
        e_eq = thermodynamics.e_eq_water_mk

    # Use short formula symbols for physical constants.
    g = constants.earth_standard_gravity
    Lv = constants.heat_of_vaporization
    Rd = constants.gas_constant_dry_air
    Rv = constants.gas_constant_water_vapor
    Cp = constants.isobaric_mass_heat_capacity

    gamma_d = g / Cp  # dry lapse rate
    w_saturated = thermodynamics.vmr2mixing_ratio(e_eq(T) / p)

    lapse = (
        gamma_d * (
            (1 + (Lv * w_saturated) / (Rd * T)) /
            (1 + (Lv**2 * w_saturated) / (Cp * Rv * T**2))
        )
    )

    return lapse


def standard_atmosphere(z, coordinates='height'):
    """International Standard Atmosphere (ISA).

    The temperature profile is defined between 0-85 km (1089 h-0.004 hPa).
    Values exceeding this range are linearly interpolated.

    Parameters:
        z (float or ndarray): Geopotential height above MSL [m]
            or pressure [Pa] (see ``coordinates``).
        coordinates (str): Either 'height' or 'pressure'.

    Returns:
        ndarray: Atmospheric temperature [K].

    Examples:

    .. plot::
        :include-source:

        import numpy as np
        from typhon.plots import (profile_p_log, profile_z)
        from typhon.physics import standard_atmosphere
        from typhon.math import nlogspace


        z = np.linspace(0, 84e3, 100)
        fig, ax = plt.subplots()
        profile_z(z, standard_atmosphere(z), ax=ax)

        p = nlogspace(1000e2, 0.4, 100)
        fig, ax = plt.subplots()
        profile_p_log(p, standard_atmosphere(p, coordinates='pressure'))

        plt.show()

    """
    h = np.array([-610, 11000, 20000, 32000, 47000, 51000, 71000, 84852])
    p = np.array(
        [108_900, 22_632, 5474.9, 868.02, 110.91, 66.939, 3.9564, 0.3734]
    )
    temp = np.array([+19.0, -56.5, -56.5, -44.5, -2.5, -2.5, -58.5, -86.28])

    if coordinates == 'height':
        z_ref = h
    elif coordinates == 'pressure':
        z_ref = np.log(p)
        z = np.log(z)
    else:
        raise ValueError(
            f'"{coordinates}" coordinate is unsupported. '
            'Use "height" or "pressure".')

    return interp1d(z_ref, temp + constants.K, fill_value='extrapolate')(z)


def pressure2height(p, T=None):
    r"""Convert pressure to height based on the hydrostatic equilibrium.

    .. math::
       z = \int -\frac{\mathrm{d}p}{\rho g}

    Parameters:
        p (ndarray): Pressure [Pa].
        T (ndarray): Temperature [K].
            If ``None`` the standard atmosphere is assumed.

    See also:
        .. autosummary::
            :nosignatures:

            standard_atmosphere

    Returns:
        ndarray: Relative height above lowest pressure level [m].
    """
    if T is None:
        T = standard_atmosphere(p, coordinates='pressure')

    layer_depth = np.diff(p)
    rho = thermodynamics.density(p, T)
    rho_layer = 0.5 * (rho[:-1] + rho[1:])

    z = np.cumsum(-layer_depth / (rho_layer * constants.g))

    return np.hstack([0, z])



# -*- coding: utf-8 -*-

"""General functions for manipulating geographical data.
"""
from numbers import Number

import imageio
import numpy as np
from sklearn.neighbors import BallTree, KDTree
from typhon.constants import earth_radius
from typhon.geodesy import geocentric2cart
from typhon.utils import split_units


__all__ = [
    'area_weighted_mean',
    'GeoIndex',
    'gridded_mean',
    'sea_mask'
]


def area_weighted_mean(lon, lat, data):
    """Calculate the mean of gridded data on a sphere.

    Data points on the Earth's surface are often represented as a grid. As the
    grid cells do not have a constant area they have to be weighted when
    calculating statistical properties (e.g. mean).

    This function returns the weighted mean assuming a perfectly spherical
    globe.

    Parameters:
        lon (ndarray): Longitude (M) angles [degree].
        lat (ndarray): Latitude (N) angles [degree].
        data ()ndarray): Data array (N x M).

    Returns:
        float: Area weighted mean.

    """
    # Calculate coordinates and steradian (in rad).
    lon = np.deg2rad(lon)
    lat = np.deg2rad(lat)
    dlon = np.diff(lon)
    dlat = np.diff(lat)

    # Longitudal mean
    middle_points = (data[:, 1:] + data[:, :-1]) / 2
    norm = np.sum(dlon)
    lon_integral = np.sum(middle_points * dlon, axis=1) / norm

    # Latitudal mean
    lon_integral *= np.cos(lat)  # Consider varying grid area (N-S).
    middle_points = (lon_integral[1:] + lon_integral[:-1]) / 2
    norm = np.sum(np.cos((lat[1:] + lat[:-1]) / 2) * dlat)

    return np.sum(middle_points * dlat) / norm


# Factor to convert a length unit to kilometers
UNITS_CONVERSION_FACTORS = [
    [{"cm", "centimeter", "centimeters"}, 1e-6],
    [{"m", "meter", "meters"}, 1e-3],
    [{"km", "kilometer", "kilometers"}, 1],
    [{"mi", "mile", "miles"}, 1.609344],  # english statute mile
    [{"yd", "yds", "yard", "yards"}, 0.9144e-3],
    [{"ft", "foot", "feet"}, 0.3048e-3],
]


def to_kilometers(distance):
    """Convert different length units to kilometers

    Args:
        distance: A string or number.

    Returns:
        A distance as float in kilometers
    """
    if isinstance(distance, Number):
        return distance
    elif not isinstance(distance, str):
        raise ValueError("Distance must be a number or a string!")

    length, unit = split_units(distance)

    if length == 0:
        raise ValueError("A valid distance length must be given!")

    if not unit:
        return length

    for units, factor in UNITS_CONVERSION_FACTORS:
        if unit in units:
            return length * factor

    raise ValueError(f"Unknown distance unit: {unit}!")


class GeoIndex:
    """Indexer that allows fast range queries with geographical coordinates"""

    def __init__(self, lat, lon, metric=None, tree_class=None,
                 shuffle=True, **tree_kwargs):
        """Initialize a GeoIndex
        
        Args:
            lat: Latitudes between -90 and 90 degrees as 1-dimensional numpy
                array.
            lon: Longitudes between -180 and 180 degrees as 1-dimensional numpy
                array. Must have the same length as `lat`.
            metric: We can use different metrics for the indexing. The default
                *minkowski* metric is the fastest but produces a big error for
                large distances. The *haversine* is more correct but is slower
                to compute. Rule of thumb: When searching for radii over
                1000km, the error is ca. 1.4 kilometers.
            tree_class: Say what tree do you want to use for building the
                spatial index. Either *Ball* or *KD* is allowed. Default is
                *Ball*.
            shuffle: The trees have a terrible performance for sorted data
                as discussed in this issue:
                https://github.com/scikit-learn/scikit-learn/issues/7687. For
                sorted, almost-gridded data (such as from SEVIRI) you should
                set this to *True*. Default is True.

        Examples:

            We have two tracks (e.g. of satellites) and we want to know whether
            and where they have been in a close distance.

            .. code-block:: python

                import matplotlib.pyplot as plt
                import numpy as np
                from typhon.geographical import GeoIndex
                from typhon.plots import worldmap

                track1 = {
                    "lat": 90*np.sin(np.linspace(-2*np.pi, 2*np.pi, 90)),
                    "lon": np.linspace(-180, 180, 90),
                }
                track2 = {
                    "lat": 90*np.cos(np.linspace(-2*np.pi, 2*np.pi, 90)),
                    "lon": np.linspace(-180, 180, 90),
                }

                # Build the index with the data of the first track:
                index = GeoIndex(track1["lat"], track1["lon"])

                # Find all points from the second track that are within a
                # radius of 500 miles to the first track
                pairs, distances = index.query(
                    track2["lat"], track2["lon"], r="500 miles"
                )

                # Plot the points
                worldmap(
                    track1["lat"], track1["lon"],
                    s=10, bg=True, label="Track 1")
                worldmap(
                    track2["lat"], track2["lon"], s=10, label="Track 2")
                worldmap(
                    track1["lat"][pairs[0]], track1["lon"][pairs[0]],
                    s=10, label="Track 1 (matched)"
                )
                worldmap(
                    track2["lat"][pairs[1]], track2["lon"][pairs[1]],
                    s=10, label="Track 2 (matched)"
                )
                plt.legend()
        """
        if metric is None:
            self.metric = "minkowski"
        else:
            self.metric = metric

        # Convert the lat and lon to the chosen metric:
        points = self._to_metric(lat, lon)

        # Save the latitudes and longitudes:
        self.lat = lat
        self.lon = lon

        if tree_class is None or tree_class == "Ball":
            tree_class = BallTree
        elif tree_class == "KD":
            tree_class = KDTree

        # KD- or ball trees have a very poor building performance for sorted
        # data (such as from SEVIRI) as discussed in this issue:
        # https://github.com/scikit-learn/scikit-learn/issues/7687
        # Hence, we shuffle the data points before inserting them.
        if shuffle:
            self.shuffler = np.arange(points.shape[0])
            np.random.shuffle(self.shuffler)

            points = points[self.shuffler]
        else:
            # The user does not want to shuffle
            self.shuffler = None

        self.tree = tree_class(
            points, **{**tree_kwargs, "metric": self.metric}
        )

    def __getitem__(self, points):
        """Get nearest indices for given points.

        Warnings:
            Not yet implemented!

        Args:
            points:

        Returns:

        """
        pass

    def _to_metric(self, lat, lon):
        if not (isinstance(lat, np.ndarray) or isinstance(lon, np.ndarray)):
            raise ValueError("lat and lon must be numpy.ndarray objects (no "
                             "pandas.Series or xarray.DataArray)!")

        if self.metric == "minkowski":
            return np.column_stack(
                geocentric2cart(earth_radius, lat, lon)
            )
        elif self.metric == "haversine":
            return np.radians(
                np.column_stack([lat, lon])
            )
        else:
            raise ValueError(f"Unknown metric '{self.metric}!'")

    def query(self, lat, lon, r, return_distance=True):
        """Find all neighbours within a radius of query points

        Args:
            lat: Latitudes between -90 and 90 degrees as 1-dimensional numpy
                array, list or single number.
            lon: Longitudes between -180 and 180 degrees as 1-dimensional numpy
                array, list or single number. Must have the same length as
                `lat`.
            r: Radius in kilometers (if number is given). You can also use
                another unit if you pass this as string (e.g. *'10 miles'*).
                Despite of the unit which is given here, the returned distances
                will always be in kilometers.
            return_distance: If True, the distances will be returned. Otherwise
                not and the query will be faster. Default is true.

        Returns:
            Two numpy arrays: *pairs* and *distances*. The first has a
            *2xN* shape, where N is the number of found collocations. The first
            row contains the indices of the points with which the GeoIndex was
            built. The second row contains the matches in the query points.
            *distances* is a numpy array with distances in kilometers.
        """
        points = self._to_metric(lat, lon)

        # The user passes the radius in kilometers but we calculate with meters
        # internally
        r = to_kilometers(r)
        if self.metric == "minkowski":
            r *= 1000.
        elif self.metric == "haversine":
            r *= 1000. / earth_radius

        results = self.tree.query_radius(
            points, r, return_distance=return_distance
        )

        if return_distance:
            jagged_pairs, jagged_distances = results
        else:
            jagged_pairs = results

        # Build the list of the collocation pairs:
        pairs = np.array([
            [build_point, query_point]
            for query_point, build_points in enumerate(jagged_pairs)
            for build_point in build_points
        ]).T

        if not return_distance:
            return pairs

        if not pairs.any():
            return pairs, pairs

        distances = np.hstack([
            distances_to_query
            for distances_to_query in jagged_distances
        ])

        # Return the distances in kilometers
        distances /= 1000.

        if self.shuffler is None:
            return pairs, distances
        else:
            # We shuffled the build points in the beginning, so the current
            # indices in the second row (the collocation indices from the build
            # points) are not correct
            pairs[0, :] = self.shuffler[pairs[0, :]]

            return pairs, distances


def gridded_mean(lat, lon, data, grid):
    """Grid data along latitudes and longitudes

    Args:
        lat: Grid points along latitudes as 1xN dimensional array.
        lon: Grid points along longitudes as 1xM dimensional array.
        data: The data as NxM numpy array.
        grid: A tuple with two numpy arrays, consisting of latitude and
            longitude grid points.

    Returns:
        Two matrices in grid form: the mean and the number of points of `data`.
    """
    grid_sum, _, _ = np.histogram2d(lat, lon, grid, weights=data)
    grid_number, _, _ = np.histogram2d(lat, lon, grid)

    return grid_sum / grid_number, grid_number


def sea_mask(lat, lon, mask):
    """Check whether geographical coordinates are over sea

    Notes:
        This uses per default a land sea mask with a grid size of 5 minutes.
        You have to decide by yourself whether it is sufficient for your data.

    Args:
        lat: Latitudes between -90 and 90 degrees as 1-dimensional numpy
                array.
        lon: Longitudes between -180 and 180 degrees as 1-dimensional numpy
            array. Must have the same length as `lat`.
        mask: Your own land-sea mask as a 2-dimensional boolean matrix or
            a path to a monochromatic PNG file. Must be sea pixels must be True
            or white, respectively.

    Returns:
        Returns a boolean array with the same dimensions as `lat`. It is True
        where the coordinate is over the sea.

    Examples:

        .. code-block:: python

            lat = np.array([60])
            lon = np.array([-0])

            is_over_sea(lat, lon, "land_water_mask_5min.png")
    """
    def _to_array(item):
        if isinstance(item, np.ndarray):
            return item

        if isinstance(item, Number):
            return np.array([item])
        else:
            return np.array(item)

    lat = _to_array(lat)
    lon = _to_array(lon)

    if not lat.size or not lon.size:
        return np.array([])

    if lon.min() < -180 or lon.max() > 180:
        raise ValueError("Longitudes out of bounds!")

    if lat.min() < -90 or lat.max() > 90:
        raise ValueError("Latitudes out of bounds!")

    if isinstance(mask, str):
        mask = np.flip(np.array(imageio.imread(mask) == 255), axis=0)

    mask_lat_step = 180 / (mask.shape[0] - 1)
    mask_lon_step = 360 / (mask.shape[1] - 1)

    lat_cell = (90 - lat) / mask_lat_step
    lon_cell = lon / mask_lon_step

    return mask[lat_cell.astype(int), lon_cell.astype(int)]





from datetime import datetime, timedelta
from os.path import dirname, join
import warnings

import numpy as np
import pandas as pd
from scipy.interpolate import interp1d
import xarray as xr

from .common import expects_file_info, HDF5


__all__ = [
    'SEVIRI',
]


class SEVIRI(HDF5):
    """File handler for SEVIRI level 1.5 HDF files
    """

    # Cache the grid of SEVIRI since it is expensive to calculate
    _grid = None

    channel_names = {
        'channel_1': 'VIS006',
        'channel_2': 'VIS008',
        'channel_3': 'IR_016',
        'channel_4': 'IR_039',
        'channel_5': 'WV_062',
        'channel_6': 'WV_073',
        'channel_7': 'IR_087',
        'channel_8': 'IR_097',
        'channel_9': 'IR_108',
        'channel_10': 'IR_120',
        'channel_11': 'IR_134',
        'channel_12': 'HRV'
    }

    def __init__(self, **kwargs):
        """

        Args:
            **kwargs: Additional key word arguments for base class.
        """

        # Call the base class initializer
        super().__init__(**kwargs)

        # We are going to import those fields per default (not channel 12
        # because it has a different resolution):
        self.standard_fields = {
            f"U-MARF/MSG/Level1.5/DATA/Channel {ch:02d}/IMAGE_DATA"
            for ch in range(1, 12)
        }

        # Standard fields can be overwritten by the user. However, we always
        # need the field with the scale and offset parameters for the channels:
        self.mandatory_fields = {
            "U-MARF/MSG/Level1.5/METADATA/HEADER/RadiometricProcessing/Level15ImageCalibration_ARRAY"  # noqa
        }

        # Map the standard fields to standard names (make also the names of all
        # dimensions more meaningful):
        self.mapping = {
            f"U-MARF/MSG/Level1.5/DATA/Channel {ch:02d}/IMAGE_DATA":
                f"channel_{ch}"
            for ch in range(1, 12)
        }
        self.reverse_mapping = {
            self.channel_names[value]: key
            for key, value in self.mapping.items()
        }
        self.mapping["U-MARF/MSG/Level1.5/METADATA/HEADER/RadiometricProcessing/Level15ImageCalibration_ARRAY"] = "counts_to_rad"  # noqa

    @staticmethod
    def grid():
        """Return the geo-grid of the SEVIRI images (lat and lon)

        Notes:
            When calling this method for the first time, the latitudes and
            longitudes are going to be calculated. This may take a while. The
            results is cached, so multiple calls are faster. The caching may
            need a lot of space (the grid is HUGE)!

        Returns:
            A xarray.Dataset with lat and lon fields
        """
        if SEVIRI._grid is not None:
            return SEVIRI._grid

        # The SEVIRI image has a size of 3712x3712 pixels
        height_width = np.arange(3712)

        # Get the intermediate coordinates
        OFF = 1856
        FAC = -781648343
        intermediate_coords = 2 ** 16 * (height_width + 1 - OFF) / FAC

        # And all their combinations
        x, y = np.meshgrid(intermediate_coords, intermediate_coords)

        # Calculate the geo locations (lat, lon) from them
        cos_x = np.cos(x)
        cos_y = np.cos(y)
        sin_y = np.sin(y)

        sa = (42164 * cos_x * cos_y) ** 2 \
            - (cos_y ** 2 + 1.006803 * sin_y ** 2) * 1737121856

        sd = np.sqrt(sa)
        sn = (42164 * cos_x * cos_y - sd) / (
                cos_y ** 2 + 1.006803 * sin_y ** 2)
        s1 = 42164 - sn * cos_x * cos_y
        s2 = sn * np.sin(x) * cos_y
        s3 = -1 * sn * sin_y
        sxy = np.sqrt(s1 ** 2 + s2 ** 2)

        # conversion to degree
        lon = np.arctan(s2 / s1) * 180 / np.pi
        lat = np.arctan(1.006803 * (s3 / sxy)) * 180 / np.pi

        # We cache the grid
        ds = xr.Dataset({
            "lat": (("line", "column"), -lat),
            "lon": (("line", "column"), -lon),
        })

        ds["lat"].attrs = {
            "long_name": "latitude",
            "units": "degrees [-90, 180]",
        }
        ds["lon"].attrs = {
            "long_name": "longitude",
            "units": "degrees [-180, 180]",
        }

        SEVIRI._grid = ds

        return SEVIRI._grid

    @expects_file_info()
    def read(self, file_info, fields=None, calibration=True, **kwargs):
        """Read SEVIRI HDF5 files and load them to a xarray.Dataset

        Args:
            file_info: Path and name of the file as string or FileInfo object.
                This can also be a tuple/list of file names or a path with
                asterisk.
            fields: Field names that you want to extract from this file as a
                list.
            **kwargs: Additional keyword arguments that are valid for
                :class:`typhon.files.handlers.common.NetCDF4`.

        Returns:
            A xrarray.Dataset object.
        """

        # Here, the user fields overwrite the standard fields:
        if fields is None:
            fields = self.standard_fields | self.mandatory_fields
        else:
            fields = self.mandatory_fields | {
                self.reverse_mapping[field]
                for field in fields
                if field not in ["time", "lat", "lon"]
            }

        # We catch the user mapping here, since we do not want to deal with
        # user-defined names in the further processing. Instead, we use our own
        # mapping at the moment, and apply the user mapping later.
        user_mapping = kwargs.pop("mapping", None)

        dataset = super(SEVIRI, self).read(
            file_info, fields=fields, mapping=self.mapping
        )

        # We set the names of each channel variable explicitly:
        for name, var in dataset.variables.items():
            if name not in ["time", "lat", "lon", "counts_to_rad"]:
                dataset[name] = ["line", "column"], var.values

        # Convert the counts to brightness temperatures:
        if calibration:
            dataset = self.counts_to_bt(dataset)

        # Add the latitudes and longitudes of the grid points:
        dataset = xr.merge([dataset, self.grid()])

        # Add the time variable (is derived from the filename normally):
        if file_info.times[0] is None:
            warnings.warn(
                "SEVIRI: The time field was not specified by the filename! Set"
                "it to 1970-01-01!"
            )
            dataset["time"] = ("time", np.array(["1970-01-01"], dtype="M8[s]"))
        else:
            dataset["time"] = ("time", [file_info.times[0]])

        # For collocating and other things, we always need the three dimensions
        # time, lat and lon to be "connected". Hence, add the time dimension to
        # each data variable as extra dimension.
        dataset = xr.concat([dataset], dim="time")

        # We want the final field names to be meaningful
        mapping = {
            old: new
            for old, new in self.channel_names.items()
            if old in dataset.variables
        }
        dataset = dataset.rename(mapping)
        return dataset

    @staticmethod
    def counts_to_bt(dataset):
        """Convert the counts to brightness temperatures

        Cold channels (4, 5, 6, 7, 8, 9, 10, 11) are converted to brightness
        temperatures while warm channels are converted to radiances.

        Args:
            dataset: A xarray.Dataset with counts

        Returns:
            A xarray.Dataset with brightness temperatures.
        """
        # We have to convert in two steps:
        # 1) Convert the counts into radiances with the given offset and scale
        # parameters (these are stored in the SEVIRI HDF file).
        # 2) Convert the radiances into brightness temperatures (only
        # applicable to the warm channels).
        conversion_table = pd.read_csv(
            join(dirname(__file__), "seviri_radiances_to_bt.csv")
        )
        for var in dataset.data_vars:
            if not var.startswith("channel_"):
                continue

            channel = int(var.split("_")[1])
            coeffs = dataset["counts_to_rad"][channel-1].item(0)
            dataset[f"channel_{channel}"] = \
                coeffs[0]*dataset[f"channel_{channel}"] + coeffs[1]

            if channel < 4:
                continue

            # Build the converter function
            converter = interp1d(
                conversion_table[f"ch{channel}"], conversion_table["BT"],
                fill_value=np.nan, bounds_error=False
            )
            dataset[f"channel_{channel}"] = xr.DataArray(
                converter(dataset[f"channel_{channel}"]),
                dims=dataset[f"channel_{channel}"].dims
            )

        # Drop the conversion variable:
        return dataset.drop("counts_to_rad")



#!/usr/bin/env python3

"""Handle specific configuration settings, such as data locations.

Configuration is handled with a configuration file with a
:mod:`configparser` syntax.

The location of the configuration file is determined by the environment
variable TYPHONRC.  If this is not set, it will use ~/.typhonrc.
"""

import pathlib
import os
from configparser import (ConfigParser, ExtendedInterpolation)


__all__ = [
    'conf',
]


conf = ConfigParser(interpolation=ExtendedInterpolation())
conf.optionxform = str
p = pathlib.Path(os.getenv("TYPHONRC", "~/.typhonrc")).expanduser()
conf.read(str(p))



"""Modules in this package contain classes to handle datasets.

That includes the overall framework to handle datasets in the dataset
module, as well as concrete datasets for specific sensors etc., including
reading routines.

To implement a new reading routine, subclass one of the datasets here.
"""

# Any commits made to this module between 2015-05-01 and 2017-03-01
# by Gerrit Holl are developed for the EC project “Fidelity and
# Uncertainty in Climate Data Records from Earth Observations (FIDUCEO)”.
# Grant agreement: 638822
# 
# All those contributions are dual-licensed under the MIT license for use
# in typhon, and the GNU General Public License version 3.

from . import dataset



__all__ = [
    'check_lat_lon',
]


def check_lat_lon(data):
    # check latitudes:
    if (data["lat"].min().item(0) < -90
            or data["lat"].max().item(0) > 90):
        values = data["lat"].min().item(0), data["lat"].max().item(0)
        raise ValueError(
            f"Latitudes are out of bounds (not in [-90, 90]): {values}"
        )

    if (data["lon"].min().item(0) < -180
            or data["lon"].max().item(0) > 180):
        values = data["lon"].min().item(0), data["lon"].max().item(0)
        raise ValueError(
            f"Longitudes are out of bounds (not in [-180, 180]): {values}"
        )


from datetime import datetime, timedelta

import numpy as np
import numexpr as ne
from netCDF4 import Dataset
from scipy.interpolate import CubicSpline
from typhon.utils import Timer
import xarray as xr

from .common import NetCDF4, expects_file_info
from .testers import check_lat_lon

__all__ = [
    'AVHRR_GAC_HDF',
    'MHS_HDF',
]


class AAPP_HDF(NetCDF4):
    """Base class for handling TOVS satellite data converted with AAPP tools

    """
    # This file handler always wants to return at least time, lat and lon
    # fields. These fields are required for this:
    standard_fields = {
        "Data/scnlintime",  # milliseconds since midnight
        "Data/scnlinyr",
        "Data/scnlindy",
        "Data/scnlin",
        "Geolocation/Latitude",
        "Geolocation/Longitude"
    }

    def __init__(self, **kwargs):
        """

        Args:
            **kwargs: Additional key word arguments for base class.
        """
        # Call the base class initializer
        super().__init__(**kwargs)

    @expects_file_info()
    def get_info(self, file_info, **kwargs):
        with Dataset(file_info.path, "r") as file:
            file_info.times[0] = \
                datetime(int(file.startdatayr[0]), 1, 1) \
                + timedelta(days=int(file.startdatady[0]) - 1) \
                + timedelta(milliseconds=int(file.startdatatime_ms[0]))
            file_info.times[1] = \
                datetime(int(file.enddatayr), 1, 1) \
                + timedelta(days=int(file.enddatady) - 1) \
                + timedelta(milliseconds=int(file.enddatatime_ms))

            return file_info

    @staticmethod
    def _get_time_field(dataset, user_fields):
        time = \
            (dataset["Data/scnlinyr"].values - 1970).astype('datetime64[Y]') \
            + (dataset["Data/scnlindy"].values - 1).astype('timedelta64[D]') \
            + dataset["Data/scnlintime"].values.astype("timedelta64[ms]")

        dataset["time"] = "scnline", time

        # Remove the time fields that we do not need any longer (expect the
        # user asked for them explicitly)
        dataset = dataset.drop(
            {"Data/scnlinyr", "Data/scnlindy", "Data/scnlintime"}
            - set(user_fields),
        )

        return dataset

    @staticmethod
    def _mask_and_scale(dataset):
        # xarray.open_dataset can mask and scale automatically, but it does not
        # know the attribute *Scale* (which is specific for AAPP files):
        for var in dataset.variables:
            # We want to remove some attributes after applying them but
            # OrderedDict does not allow to pop the values:
            attrs = dict(dataset[var].attrs)

            mask = attrs.pop('FillValue', None)
            if mask is not None:
                dataset[var] = dataset[var].where(
                    # Also cover overflow errors as they are in
                    # NSS.MHSX.NN.D07045.S2234.E0021.B0896162.GC.h5
                    (dataset[var] != mask) & (dataset[var] != -2147483648.0)
                )

            scaling = attrs.pop('Scale', None)
            if scaling is not None:
                dataset[var] = dataset[var].astype(float) * scaling

            dataset[var].attrs = attrs

    def _test_coords(self, dataset, wanted=None):
        # Maximal these dimensions (or less) should be in the dataset:
        if wanted is None:
            wanted = {'channel', 'scnline', 'scnpos'}
        reality = set(dataset.dims.keys())

        if reality - wanted:
            raise ValueError(
                f"Unexpected dimension in AAPP file! {reality - wanted}"
            )


class MHS_HDF(AAPP_HDF):
    """File handler for MHS level 1C HDF files
    """
    def __init__(self, **kwargs):
        super(MHS_HDF, self).__init__(**kwargs)

        # Map the standard fields to standard names (make also the names of all
        # dimensions more meaningful):
        self.mapping = {
            "Geolocation/Latitude": "lat",
            "Geolocation/Longitude": "lon",
            "Data/scnlin": "scnline",
            "Data/phony_dim_0": "scnline",
            "Data/phony_dim_1": "scnpos",
            "Data/phony_dim_2": "channel",
            "Geolocation/phony_dim_3": "scnline",
            "Geolocation/phony_dim_4": "scnpos",
        }

    @expects_file_info()
    def read(self, file_info, mask_and_scale=True, **kwargs):
        """Read and parse MHS AAPP HDF5 files and load them to xarray

        Args:
            file_info: Path and name of the file as string or FileInfo object.
                This can also be a tuple/list of file names or a path with
                asterisk.
            mask_and_scale: Where the data contains missing values, it will be
                masked with NaNs. Furthermore, data with scaling attributes
                will be scaled with them.
            **kwargs: Additional keyword arguments that are valid for
                :class:`~typhon.files.handlers.common.NetCDF4`.

        Returns:
            A xrarray.Dataset object.
        """

        # Make sure that the standard fields are always gonna be imported:
        user_fields = kwargs.pop("fields", {})
        if user_fields:
            fields = self.standard_fields | set(user_fields)
        else:
            fields = None

        # We catch the user mapping here, since we do not want to deal with
        # user-defined names in the further processing. Instead, we use our own
        # mapping
        user_mapping = kwargs.pop("mapping", None)

        # Load the dataset from the file:
        dataset = super().read(
            file_info, fields=fields, mapping=self.mapping,
            mask_and_scale=mask_and_scale, **kwargs
        )

        scnlines = dataset["scnline"].values
        dataset = dataset.assign_coords(
             scnline=dataset["scnline"]
        )

        dataset["scnline"] = np.arange(1, dataset.scnline.size+1)
        dataset["scnpos"] = np.arange(1, 91)
        dataset["channel"] = "channel", np.arange(1, 6)

        # Create the time variable (is built from several other variables):
        dataset = self._get_time_field(dataset, user_fields)

        if mask_and_scale:
            self._mask_and_scale(dataset)

        # Make a fast check whether everything is alright
        self._test_coords(dataset)

        # Check the latitudes and longitudes:
        check_lat_lon(dataset)

        if user_mapping is not None:
            dataset = dataset.rename(user_mapping)

        return dataset


class AVHRR_GAC_HDF(AAPP_HDF):
    """File handler for AVHRR GAC level 1C HDF files
    """

    def __init__(self, **kwargs):
        super(AVHRR_GAC_HDF, self).__init__(**kwargs)

        # Map the standard fields to standard names (make also the names of all
        # dimensions more meaningful):
        self.mapping = {
            "Geolocation/Latitude": "lat",
            "Geolocation/Longitude": "lon",
            "Data/scnlin": "scnline",
            "Data/phony_dim_0": "scnline",
            "Data/phony_dim_1": "scnpos",
            "Data/phony_dim_2": "channel",
            "Data/phony_dim_3": "calib",
            "Geolocation/phony_dim_4": "scnline",
            "Geolocation/phony_dim_5": "packed_pixels",
        }

    @expects_file_info()
    def read(self, file_info, mask_and_scale=True, interpolate_packed_pixels=True,
             max_nans_interpolation=10, **kwargs):
        """Read and parse MHS AAPP HDF5 files and load them to xarray

        Args:
            file_info: Path and name of the file as string or FileInfo object.
                This can also be a tuple/list of file names or a path with
                asterisk.
            mask_and_scale: Where the data contains missing values, it will be
                masked with NaNs. Furthermore, data with scaling attributes
                will be scaled with them.
            interpolate_packed_pixels: Geo-location data is packed and must be
                interpolated to use them as reference for each pixel.
            max_nans_interpolation: How many NaN values are allowed in latitude
                and longitudes before raising an error?
            **kwargs: Additional keyword arguments that are valid for
                :class:`~typhon.files.handlers.common.NetCDF4`.

        Returns:
            A xrarray.Dataset object.
        """

        # Make sure that the standard fields are always gonna be imported:
        user_fields = kwargs.pop("fields", {})
        if user_fields:
            fields = self.standard_fields | set(user_fields)
        else:
            fields = None

        # We catch the user mapping here, since we do not want to deal with
        # user-defined names in the further processing. Instead, we use our own
        # mapping
        user_mapping = kwargs.pop("mapping", None)

        # Load the dataset from the file:
        dataset = super().read(
            file_info, fields=fields, mapping=self.mapping,
            mask_and_scale=mask_and_scale, **kwargs
        )

        # Keep the original scnlines
        scnlines = dataset["scnline"].values

        dataset = dataset.assign_coords(
            scnline=dataset["scnline"]
        )

        dataset["scnline"] = np.arange(1, dataset.scnline.size+1)
        dataset["scnpos"] = np.arange(1, 2049)
        dataset["channel"] = "channel", np.arange(1, 6)

        # Currently, the AAPP converting tool seems to have a bug. Instead of
        # retrieving 409 pixels per scanline, one gets 2048 pixels. The
        # additional values are simply duplicates (or rather quintuplicates):
        dataset = dataset.sel(scnpos=slice(4, None, 5))
        dataset["scnpos"] = np.arange(1, 410)

        # Create the time variable (is built from several other variables):
        dataset = self._get_time_field(dataset, user_fields)

        if mask_and_scale:
            self._mask_and_scale(dataset)

        # All geolocation fields are packed in the AVHRR GAC files:
        if interpolate_packed_pixels:
            self._interpolate_packed_pixels(dataset, max_nans_interpolation)
            allowed_coords = {'channel', 'calib', 'scnline', 'scnpos'}
        else:
            allowed_coords = {'channel', 'calib', 'scnline', 'scnpos',
                              'packed_pixels'}

        # Make a fast check whether everything is alright
        self._test_coords(dataset, allowed_coords)

        # Check the latitudes and longitudes:
        check_lat_lon(dataset)

        if user_mapping is not None:
            dataset = dataset.rename(user_mapping)

        return dataset

    @staticmethod
    def _interpolate_packed_pixels(dataset, max_nans_interpolation):
        given_pos = np.arange(5, 409, 8)
        new_pos = np.arange(1, 410)

        lat_in = np.deg2rad(dataset["lat"].values)
        lon_in = np.deg2rad(dataset["lon"].values)

        # We cannot define given positions for each scanline, but we have to
        # set them for all equally. Hence, we skip every scan position of all
        # scan lines even if only one contains a NaN value:
        nan_scnpos = \
            np.isnan(lat_in).sum(axis=0) + np.isnan(lon_in).sum(axis=0)
        valid_pos = nan_scnpos == 0

        if valid_pos.sum() < 52 - max_nans_interpolation:
            raise ValueError(
                "Too many NaNs in latitude and longitude of this AVHRR file. "
                "Cannot guarantee a good interpolation!"
            )

        # Filter NaNs because CubicSpline cannot handle it:
        lat_in = lat_in[:, valid_pos]
        lon_in = lon_in[:, valid_pos]
        given_pos = given_pos[valid_pos]

        x_in = np.cos(lon_in) * np.cos(lat_in)
        y_in = np.sin(lon_in) * np.cos(lat_in)
        z_in = np.sin(lat_in)

        xf = CubicSpline(given_pos, x_in, axis=1, extrapolate=True)(new_pos)
        yf = CubicSpline(given_pos, y_in, axis=1, extrapolate=True)(new_pos)
        zf = CubicSpline(given_pos, z_in, axis=1, extrapolate=True)(new_pos)
        lon = np.rad2deg(np.arctan2(yf, xf))
        lat = np.rad2deg(np.arctan2(zf, np.sqrt(xf ** 2 + yf ** 2)))

        dataset["lat"] = ("scnline", "scnpos"), lat
        dataset["lon"] = ("scnline", "scnpos"), lon

        # The other packed variables will be simply padded:
        for var_name, var in dataset.data_vars.items():
            if "packed_pixels" not in var.dims:
                continue

            nan_scnpos = np.isnan(var).sum(axis=0)
            valid_pos = nan_scnpos == 0
            given_pos = np.arange(5, 409, 8)[valid_pos]

            dataset[var_name] = xr.DataArray(
                CubicSpline(
                    given_pos, var.values[:, valid_pos], axis=1,
                    extrapolate=True)(new_pos),
                dims=("scnline", "scnpos")
            )



r"""
This subpackage provides functionality to sample posterior distributions of
inverse problems in atmospheric soundings using ARTS as forward model.

The main functionality is implemented by the `MCMC` class which implements
the Metropolis algorithm to sample from the posterior distribution.

In addition to that this subpackage provides a `RandomWalk` class that
simplifies the setup of random walk jump functions as well as diagnostic
function to assess mixing and convergence of the simulations.
"""
from typhon.retrieval.mcmc.mcmc import MCMC, r_factor, autocorrelation, \
                                        split, effective_sample_size
from typhon.retrieval.mcmc.jumping_rules import RandomWalk



"""
Module that provides classes for tree creation and handling.

Trees are powerful structures to sort a huge amount of data and to speed up
performing query requests on them significantly.
"""

from collections import Iterable

import pandas as pd
import numba
import numpy as np

from sklearn.neighbors import BallTree, KDTree

__all__ = [
    "IntervalTree",
    "RangeTree",
]


class IntervalTreeNode:
    """Helper class for IntervalTree.

    """
    def __init__(self, center_point, center, left, right):
        self.center_point = center_point
        self.center = np.asarray(center)
        self.left = left
        self.right = right


class IntervalTree:
    """Tree to implement fast 1-dimensional interval searches.

    Based on the description in Wikipedia
    (https://en.wikipedia.org/wiki/Interval_tree#Centered_interval_tree)
    and the GitHub repository by tylerkahn
    (https://github.com/tylerkahn/intervaltree-python).

    Examples:
        Check 1000 intervals on 1000 other intervals:

    .. code-block:: python

        import numpy as np
        from typhon.trees import IntervalTree

        intervals = np.asarray([np.arange(1000)-0.5, np.arange(1000)+0.5]).T
        tree = IntervalTree(intervals)
        query_intervals = [[i-1, i+1] for i in range(1000)]
        results = tree.query(query_intervals)

    """
    def __init__(self, intervals):
        """Creates an IntervalTree object.

        Args:
            intervals: A list or numpy.array containing the intervals (list of
                two numbers).
        """
        if not isinstance(intervals, np.ndarray):
            intervals = np.asarray(intervals)

        # Check the intervals whether they are valid:
        self.left = np.min(intervals)
        self.right = np.max(intervals)

        # We want to return the indices of the intervals instead of their
        # actual bounds. But the original indices will be lost due resorting.
        # Hence, we add the original indices to the intervals themselves.
        indices = np.arange(intervals.shape[0]).reshape(intervals.shape[0], 1)
        indexed_intervals = np.hstack([intervals, indices])
        self.root = self._build_tree(np.sort(indexed_intervals, axis=0))

    def __contains__(self, item):
        if isinstance(item, (tuple, list)):
            return bool(self._query(item, self.root, check_extreme=True))
        else:
            return bool(self._query_point(item, self.root, check_extreme=True))

    # @numba.jit
    def _build_tree(self, intervals):
        if not intervals.any():
            return None

        center_point = self._get_center(intervals)

        # Sort the intervals into bins
        center = intervals[(intervals[:, 0] <= center_point)
                           & (intervals[:, 1] >= center_point)]
        left = intervals[intervals[:, 1] < center_point]
        right = intervals[intervals[:, 0] > center_point]

        return IntervalTreeNode(
            center_point, center,
            self._build_tree(left), self._build_tree(right)
        )

    @staticmethod
    def _get_center(intervals):
        return intervals[int(intervals.shape[0]/2), 0]

    @staticmethod
    def interval_overlaps(interval1, interval2):
        """Checks whether two intervals overlap each other.

        Args:
            interval1: A tuple of two numbers: the lower and higher bound of
                the first interval.
            interval2: A tuple of two numbers: the lower and higher bound of
                the second interval.

        Returns:
            True if the intervals overlap.
        """
        return interval1[0] <= interval2[1] and interval1[1] >= interval2[0]

    @staticmethod
    def interval_contains(interval, point):
        """Checks whether a point lies in a interval.

        Args:
            interval: A tuple of two numbers: the lower and higher bound of the
                first interval.
            point: The point (just a number)

        Returns:
            True if point lies in the interval.
        """
        return interval[0] <= point <= interval[1]

    # @numba.jit
    def query(self, intervals):
        """Find all overlaps between this tree and a list of intervals.

        Args:
            intervals: A list of intervals. Each interval is a tuple/list of
                two elements: its lower and higher boundary.

        Returns:
            List of lists which contain the overlapping intervals of this tree
            for each element in `intervals`.
        """
        return [self._query(interval, self.root, check_extreme=True)
                for interval in intervals]

    # @numba.jit
    def _query(self, query_interval, node, check_extreme=False):
        # Check this special case: the bounds of the query interval lie outside
        # of the bounds of this tree:
        if (check_extreme
                and IntervalTree.interval_contains(query_interval, self.left)
                and IntervalTree.interval_contains(query_interval, self.right)):
            return []  # TODO: Return all intervals

        # Let's start with the centered intervals
        intervals = [int(interval[2]) for interval in node.center
                     if IntervalTree.interval_overlaps(interval, query_interval)]

        if query_interval[0] <= node.center_point and node.left is not None:
            intervals.extend(self._query(query_interval, node.left))

        if query_interval[1] >= node.center_point and node.right is not None:
            intervals.extend(self._query(query_interval, node.right))

        return intervals

    # @numba.jit
    def query_points(self, points):
        """Find all intervals of this tree which contain one of those points.

        Args:
            points: A list of points.

        Returns:
            List of lists which contain the enclosing intervals of this tree
            for each element in `points`.
        """
        return [self._query_point(point, self.root, check_extreme=True)
                for point in points]

    # @numba.jit
    def _query_point(self, point, node, check_extreme=False):
        # Check this special case: the query point lies outside of the bounds
        # of this tree:
        if check_extreme \
                and not IntervalTree.interval_contains(
                    (self.left, self.right), point):
            return []

        # Let's start with the centered intervals
        intervals = [int(interval[2])
                     for interval in node.center
                     if IntervalTree.interval_contains(interval, point)]

        if point < node.center_point and node.left is not None:
            intervals.extend(self._query_point(point, node))

        if point > node.center_point and node.right is not None:
            intervals.extend(self._query_point(point, node))

        return intervals


class RangeTree:
    def __init__(self, points, shuffle=True, tree_class=None):
        # KD- or ball trees have a very poor building performance for sorted
        # data (such as from SEVIRI) as discussed in this issue:
        # https://github.com/scikit-learn/scikit-learn/issues/7687
        # Hence, we shuffle the data points before inserting them.
        if shuffle:
            self.shuffler = np.arange(points.shape[0])
            np.random.shuffle(self.shuffler)

            points = points[self.shuffler]
        else:
            # The user does not want to shuffle
            self.shuffler = None

        if tree_class is None or tree_class == "Ball":
            tree_class = BallTree
        elif tree_class == "KD":
            tree_class = KDTree

        tree_points = np.column_stack([points, np.zeros_like(points)])
        self.tree = tree_class(tree_points)

    def query_radius(self, points, r):
        query_points = np.column_stack([points, np.zeros_like(points)])

        jagged_pairs = self.tree.query_radius(query_points, r)

        # Build the list of pairs:
        pairs = np.array([
            [build_point, query_point]
            for query_point, build_points in enumerate(jagged_pairs)
            for build_point in build_points
        ]).T

        if self.shuffler is None:
            return pairs
        else:

            # We shuffled the build points in the beginning, so the current
            # indices in the second row (the collocation indices from the build
            # points) are not correct
            pairs[0, :] = self.shuffler[pairs[0, :]]

            return pairs



# -*- coding: utf-8 -*-

"""This module contains convenience functions for any purposes.
"""
from typhon.utils.cache import *  # noqa
from typhon.utils.common import *  # noqa
from typhon.utils.sphinxext import *  # noqa
from typhon.utils.timeutils import *  # noqa


__all__ = [s for s in dir() if not s.startswith('_')]



from collections import defaultdict
from copy import copy
from datetime import datetime
from functools import wraps
import glob
from inspect import signature, ismethod
import os
import pickle
import warnings

import netCDF4
import pandas as pd
import xarray as xr
import numpy as np

# The HDF4 file handler needs pyhdf, this might be very tricky to install if
# you cannot use anaconda. Hence, I do not want it to be a hard dependency:
pyhdf_is_installed = False
try:
    from pyhdf import HDF, VS, V
    from pyhdf.SD import SD, SDC
    pyhdf_is_installed = True
except ImportError:
    pass

# The HDF5 file handler needs h5py, this might be very tricky to install if
# you cannot use anaconda. Hence, I do not want it to be a hard dependency:
h5py_is_installed = False
try:
    import h5py
    h5py_is_installed = True
except ImportError:
    pass

__all__ = [
    'CSV',
    'FileHandler',
    'FileInfo',
    'HDF4',
    'HDF5',
    'NetCDF4',
    'Plotter',
    'expects_file_info',
    # 'Numpy',
    # 'Pickle',
    # 'XML'
]


def parametrized(dec):
    """A decorator for decorators that need parameters

    Do not think about this too long, it may cause headaches. Have a look at
    this instead: https://stackoverflow.com/a/26151604

    Args:
        dec: A decorator function

    Returns:
        The decoratored decorator function.
    """
    def layer(*args, **kwargs):
        def repl(f):
            return dec(f, *args, **kwargs)
        return repl
    return layer


@parametrized
def expects_file_info(method, pos=None, key=None):
    """Convert a method argument to a :class:`FileInfo` object

    This is a decorator function that can take parameters.

    If the argument is already a FileInfo object, nothing happens.

    Args:
        method: Method object that should be decorated.
        pos: The index of the file info in the positional argument list.
            Default is 1 (assumes to decorate a method).
        key: The key of the file info in the key word argument dict.

    Returns:
        The return value of the decorated method.

    Examples:

        .. code-block:: python

        @expects_file_info()(0)
        def read(file, *args, *kwargs):
            # file is a Fileinfo object now with the attribute path containing
            # "path/to/file.txt"

        read("path/to/file.txt")

    """

    if pos is None and key is None:
        pos = 1

    @wraps(method)
    def wrapper(*args, **kwargs):
        args = list(args)
        if args and pos is not None:
            if not isinstance(args[pos], FileInfo):
                args[pos] = FileInfo(args[pos])
        else:
            if not isinstance(kwargs[key], FileInfo):
                kwargs[key] = FileInfo(kwargs[key])

        return method(*args, **kwargs)
    return wrapper


def _xarray_rename_fields(dataset, mapping):
    if mapping is not None:
        # Maybe some variables should be renamed that are not in the
        # dataset any longer?
        names = set(dataset.dims.keys()) | set(dataset.variables.keys())

        mapping = {
            old_name: new_name
            for old_name, new_name in mapping.items()
            if old_name in names
        }

        dataset = dataset.rename(mapping)

    return dataset


class FileHandler:
    """Base file handler class.

    This can be used alone or with the :class:`~typhon.files.fileset.FileSet`
    classes. You can either initialize specific *reader* ,*info* or *writer*
    functions or you can inherit from this class and override its methods. If
    you need a very specialised and reusable file handler class, you should
    consider following the second approach.
    """

    def __init__(
            self, reader=None, info=None, writer=None, **kwargs):
        """Initialize a filer handler object.

        Args:
            reader: Reference to a function that defines how to read a given
                file and returns an object with the read data. The function
                must accept a :class:`FileInfo` object as first parameter.
            info: Reference to a function that returns a :class:`FileInfo`
                object with information about the given file. You cannot use
                the :meth:`get_info` without setting this parameter. The
                function must accept a filename as string as first parameter.
            writer: Reference to a function that defines how to write the data
                to a file. The function must accept the data object as first
                and a :class:`FileInfo` object as second parameter.
        """

        self.reader = reader
        self.info = info
        self.writer = writer

        # If you want to ravel / flat the data coming from this file handler
        # (e.g., this is necessary for collocation routines), you need the
        # dimension names that you can stack on top each other.
        self.stack_dims = {}

    @expects_file_info()
    def get_info(self, filename, **kwargs):
        """Return a :class:`FileInfo` object with parameters about the
        file content.

        Notes:
            This is the base class method that does nothing per default.

        Args:
            filename: A string containing path and name or a :class:`FileInfo`
                object of the file of which to get the information about.
            **kwargs: Additional keyword arguments.

        Returns:
            A :class:`FileInfo` object.
        """
        if self.info is not None:
            # Some functions do not accept additional key word arguments (via
            # kwargs). And if they are methods, they accept an additional
            # "self" or "class" parameter.
            number_args = 1 + int(ismethod(self.info))
            if len(signature(self.info).parameters) > number_args:
                return self.info(filename, **kwargs)
            else:
                return self.info(filename)

        raise NotImplementedError(
            "This file handler does not support reading data from a file. You "
            "should use a different file handler.")

    @expects_file_info()
    def read(self, filename, **kwargs):
        """Open a file by its name, read its content and return it

        Notes:
            This is the base class method that does nothing per default.

        Args:
            filename: A string containing path and name or a :class:`FileInfo`
                object of the file from which to read.
            **kwargs: Additional key word arguments.

        Returns:
            An object containing the file's content (e.g. numpy array, etc.).
        """
        if self.reader is not None:
            # Some functions do not accept additional key word arguments (via
            # kwargs). And if they are methods, they accept an additional
            # "self" or "class" parameter.
            number_args = 1 + int(ismethod(self.reader))
            if len(signature(self.reader).parameters) > number_args:
                return self.reader(filename, **kwargs)
            else:
                return self.reader(filename)

        raise NotImplementedError(
            "This file handler does not support reading data from a file. You "
            "should use a different file handler.")

    @expects_file_info(pos=2)
    def write(self, data, filename, **kwargs):
        """Store a data object to a file.

        Notes:
            This is the base class method that does nothing per default.

        Args:
            filename: A string containing path and name or a :class:`FileInfo`
                object to which to store the data. Existing files will be
                overwritten.
            data: Object with data (e.g. numpy array, etc.).

        Returns:
            None
        """
        if self.writer is not None:
            if len(signature(self.writer).parameters) > 2:
                self.writer(data, filename, **kwargs)
            else:
                self.writer(data, filename)

            return None

        raise NotImplementedError(
            "This file handler does not support writing data to a file. You "
            "should use a different file handler.")


class FileInfo(os.PathLike):
    """Container of information about a file (time coverage, etc.)

    This is a simple object that holds the path and name, time coverage and
    further attributes of a file. It fulfills the os.PathLike protocol, i.e.
    you can use it as filename argument for the most python functions.

    See this Example:

    .. code-block:: python

        # Initialise a FileInfo object that points to a file
        file_info = FileInfo(
            path="path/to/a/file.txt",
            # The time coverage of the file (needed by Dataset classes)
            times=[datetime(2018, 1, 1), datetime(2018, 1, 10)],
            # Additional attributes:
            attr={},
        )

        with open(file_info) as file:
            ...

        # If you need to access the path or other attributes directly, you can
        # do it like this:
        file_info.path   # "path/to/a/file.txt"
        file_info.times  # [datetime(2018, 1, 1), datetime(2018, 1, 10)]
        file_info.attr   # {}
    """
    def __init__(self, path=None, times=None, attr=None):
        """Initialise a FileInfo object.

        Args:
            path: Absolute path to a file.
            times: A list or tuple of two datetime objects indicating start and
                end time of the file.
            attr: A dictionary with further attributes.
        """
        super(FileInfo, self).__init__()

        self._path = None
        self.path = path

        self._times = None
        self.times = times

        if attr is None:
            self.attr = {}
        else:
            self.attr = attr

    def __eq__(self, other):
        return self.path == other.path and self.times == other.times

    def __fspath__(self):
        return self.path

    def __hash__(self):
        # With this we can use this FileInfo object also as a key in a
        # dictionary
        return hash(self.path)

    def __repr__(self):
        return f"FileInfo(\n  '{self.path}',\n" \
               f"  times={self.times},\n" \
               f"  attr={self.attr}\n)"

    def __str__(self):
        return self.path

    def copy(self):
        return copy(self)

    @classmethod
    def from_json_dict(cls, json_dict):
        times = []
        for i in range(2):
            if json_dict["times"][i] is None:
                times.append([None])
            else:
                times.append(
                    datetime.strptime(
                        json_dict["times"][i], "%Y-%m-%dT%H:%M:%S.%f"),
                )

        return cls(json_dict["path"], times, json_dict["attr"])

    @property
    def path(self):
        return self._path

    @path.setter
    def path(self, value):
        if isinstance(value, FileInfo):
            raise ValueError("You cannot set path to a FileInfo object.")
        self._path = value

    @property
    def times(self):
        return self._times

    @times.setter
    def times(self, value):
        if value is None:
            self._times = [None, None]
        else:
            self._times = list(value)
            if len(self._times) != 2:
                raise ValueError("FileInfo.times can only be a list of two "
                                 "timestamps!")

    def update(self, other_info, ignore_none_time=True):
        """Update this object with another FileInfo object.

        Args:
            other_info: A FileInfo object.
            ignore_none_time: If the start time or end time of *other_info* is
                set to None, it does not overwrite the corresponding time of
                this object.

        Returns:
            None
        """
        self.attr.update(**other_info.attr)

        if other_info.times[0] is not None or not ignore_none_time:
            self.times[0] = other_info.times[0]
        if other_info.times[1] is not None or not ignore_none_time:
            self.times[1] = other_info.times[1]

    def to_json_dict(self):
        return {
            "path": self.path,
            "times": [
                self.times[0].strftime("%Y-%m-%dT%H:%M:%S.%f"),
                self.times[1].strftime("%Y-%m-%dT%H:%M:%S.%f")
            ],
            "attr": self.attr,
        }


class CSV(FileHandler):
    """File handler that can read / write data from / to a CSV file

    A CSV file is file containing data separated by commas (or by any other
    delimiter).
    """
    def __init__(self, info=None):
        """Initializes a CSV file handler class.

        Args:
            info: A function that returns a :class:`FileInfo` object of a
                given file.
        """
        # Call the base class initializer
        super().__init__(info=info)

    @expects_file_info()
    def read(self, file_info, fields=None, **kwargs):
        """Read a CSV file and return an xarray.Dataset with its content

        Args:
            file_info: Path and name of the file as string or FileInfo object.
            fields: Field that you want to extract from the file. If not given,
                all fields are going to be extracted.
            **kwargs: Additional keyword arguments for the pandas function
                `pandas.read_csv`. See for more details:
                https://pandas.pydata.org/pandas-docs/stable/generated/pandas.read_csv.html

        Returns:
            A xarray.Dataset object.
        """

        data = pd.read_csv(file_info.path, **kwargs).to_xarray()

        if fields is None:
            return data
        else:
            return data[fields]

    @expects_file_info(pos=2)
    def write(self, data, file_info, **kwargs):
        """Write a xarray.Dataset to a CSV file.

        Args:
            data: An DataGroup object that should be saved.
            file_info: Path and name of the file as string or FileInfo object.
            **kwargs: Additional keyword arguments for
                `pandas.Dataframe.to_csv`. See for more details:
                https://pandas.pydata.org/pandas-docs/stable/generated/pandas.DataFrame.to_csv.html

        Returns:
            None
        """
        data.to_dataframe().to_csv(file_info.path, **kwargs)


class HDF4(FileHandler):
    """File handler that can read data from a HDF4 file
    """
    def __init__(self, info=None):
        """Initializes a CSV file handler class.

        Args:
            info: A function that returns a :class:`FileInfo` object of a
                given file.
        """
        if not pyhdf_is_installed:
            raise ImportError("Could not import pyhdf, which is necessary for "
                              "reading HDF4 files!")

        # Call the base class initializer
        super().__init__(info=info)

    @expects_file_info()
    def read(self, file_info, fields=None, mapping=None):
        """Read and parse HDF4 files and load them to a xarray.Dataset

        Args:
            file_info: Path and name of the file as string or FileInfo object.
            fields: Field names that you want to extract from this file as a
                list.
            mapping: A dictionary that maps old field names to new field names.
                If given, `fields` must contain the old field names.

        Returns:
            A xarray.Dataset object.
        """

        if fields is None:
            raise NotImplementedError(
                "You have to set field names. Loading the complete file is not"
                " yet implemented!"
            )

        dataset = xr.Dataset()

        # Files in HDF4 format are not very pretty. This code is taken from
        # http://hdfeos.org/zoo/OTHER/2010128055614_21420_CS_2B-GEOPROF_GRANULE_P_R04_E03.hdf.py
        # and adapted by John Mrziglod.

        file = HDF.HDF(file_info.path)

        try:
            vs = file.vstart()

            for field in fields:
                # Add the field data to the dataset.
                dataset[field] = self._get_field(vs, field)
        except Exception as e:
            raise e
        finally:
            file.close()

        return _xarray_rename_fields(dataset, mapping)

    @staticmethod
    def _get_field(vs, field):
        field_id = vs.find(field)

        if field_id == 0:
            # Field was not found.
            warnings.warn(
                "Field '{0}' was not found!".format(field), RuntimeWarning
            )

        field_id = vs.attach(field_id)
        nrecs, _, _, _, _ = field_id.inquire()
        raw_data = field_id.read(nRec=nrecs)
        data = xr.DataArray(raw_data).squeeze()
        field_id.detach()
        return data


class HDF5(FileHandler):
    """File handler for SEVIRI level 1.5 HDF files
    """

    def __init__(self, **kwargs):
        """

        Args:
            **kwargs: Additional key word arguments for base class.
        """
        if not h5py_is_installed:
            raise ImportError("Could not import h5py, which is necessary for "
                              "reading HDF5 files!")

        # Call the base class initializer
        super().__init__(**kwargs)

    @expects_file_info()
    def read(self, file_info, fields=None, mapping=None, **kwargs):
        """Read SEVIRI HDF5 files and load them to a xarray.Dataset

        Args:
            file_info: Path and name of the file as string or FileInfo object.
                This can also be a tuple/list of file names or a path with
                asterisk.
            fields: ...
            **kwargs: Additional keyword arguments that are valid for
                :class:`typhon.files.handlers.common.NetCDF4`.

        Returns:
            A xrarray.Dataset object.
        """

        # Here, the user fields overwrite the standard fields:
        if fields is None:
            raise NotImplementedError(
                "Loading complete HDF5 files without giving explicit field "
                "names is not yet implemented!"
            )

        # keys are dimension size, values are dimension names
        dim_dict = {}

        # Load the dataset from the file:
        with h5py.File(file_info.path, 'r') as file:
            dataset = xr.Dataset()

            for field in fields:
                if field not in file:
                    raise KeyError(f"No field named '{field}'!")

                dims = []
                for dim_size in file[field].shape:
                    dim_name = dim_dict.get(dim_size, None)
                    if dim_name is None:
                        dim_name = f"dim_{len(dim_dict)}"
                        dim_dict[dim_size] = dim_name

                    dims.append(dim_name)

                dataset[field] = xr.DataArray(
                    file[field], dims=dims,
                    # Currently, some attributes may contain byte-strings that
                    # are not nice for further processing
                    attrs={}, #dict(file[field].attrs)
                )

            xr.decode_cf(dataset, **kwargs)
            dataset.load()

        return _xarray_rename_fields(dataset, mapping)


class NetCDF4(FileHandler):
    """File handler that can load / store xarray.Dataset from / to NetCDF4

    This file handler can also handle pseudo groups in :class:`xarray.Dataset`
    objects.
    """

    def __init__(self, **kwargs):
        """Initialize a NetCDF4 file handler class

        Args:
            info: You cannot use the :meth:`get_info` without giving a
                function here that returns a FileInfo object.
        """
        # Call the base class initializer
        super().__init__(**kwargs)

    @expects_file_info()
    def read(self, file_info, fields=None, mapping=None, **kwargs):
        """Read and parse NetCDF files and load them to a xarray.Dataset

        Args:
            file_info: Path and name of the file as string or FileInfo object.
                This can also be a tuple/list of file names or a path with
                asterisk (this is still not implemented!).
            fields: List of field names that should be read. The other fields
                will be ignored. If `mapping` is given, this should contain the
                new field names.
            mapping: A dictionary which is used for renaming the fields. If
                given, `fields` must contain the old field names.
            **kwargs: Additional keyword arguments for
                :func:`xarray.decode_cf` such as `mask_and_scale`, etc.

        Returns:
            A xarray.Dataset object.

        Examples:

            .. code-block:: python

                from typhon.files import NetCDF4

                fh = NetCDF4()
                data = fh.read("filename.nc")

                # OR if you want to load only some fields:
                data = fh.read("filename.nc", fields=["temp", "lat", "lon"])

        """
        # xr.open_dataset does still not support loading all groups from a
        # file except a very cumbersome (and expensive) way by using the
        # parameter `group`. To avoid this, we load all groups and their
        # variables by using the netCDF4 directly and load them later into a
        # xarray dataset.

        with netCDF4.Dataset(file_info.path, "r") as root:
            # xarray decode_cf scales, don't do it twice!
            root.set_auto_scale(False)
            dataset = xr.Dataset()
            self._load_group(dataset, None, root, fields)

            dataset = xr.decode_cf(dataset, **kwargs)

        return _xarray_rename_fields(dataset, mapping)

    @staticmethod
    def _get_dimension_name(ds, group, path, dim):
        # If the dimension is defined in the subgroup, use NOT the one of the
        # parent group:
        if dim in group.variables or path == "":
            # use the subgroup dimension!
            return path + dim

        # Go through all ancestor groups (start with the parent, then
        # grandparent, etc.) and check whether there is a dimension that suits:
        ancestor_groups = [None] + path[:-1].split("/")[:-1]

        for i, ancestor_group in enumerate(reversed(ancestor_groups)):
            if ancestor_group is None:
                ancestor_dim = dim
            else:
                ancestor_dim = "/".join(
                    ancestor_groups[1:len(ancestor_groups) - i] + [dim])

            ancestor_size = ds.dims.get(ancestor_dim, None)

            if ancestor_size is not None \
                    and group.dimensions[dim].size == ancestor_size:
                # use the ancestor dimension:
                return ancestor_dim

        # use the subgroup dimension!
        return path + dim

    @staticmethod
    def _load_group(ds, path, group, fields):
        if path is None:
            # The current group is the root group
            path = ""
            ds.attrs = dict(group.__dict__)
        else:
            path += "/"

        # Dimension (coordinate) mapping: A dimension might be defined in a
        # group, then it is valid for this group only. Otherwise, the
        # dimension from the parent group is taken (if it suits with name and
        # size)
        dim_map = {
            dim: NetCDF4._get_dimension_name(ds, group, path, dim)
            for dim in group.dimensions
        }

        # Load variables:
        try:
            for var_name, var in group.variables.items():
                if fields is None or path + var_name in fields:
                    dims = [dim_map[dim] for dim in var.dimensions]
                    if len(dims) == 0 and var[:] is np.ma.masked:
                        ds[path + var_name] = dims, np.nan, dict(var.__dict__)
                    else:
                        ds[path + var_name] = dims, var[:], dict(var.__dict__)
        except RuntimeError:
            raise KeyError(f"Could not load the variable {path + var_name}!")

        # Do the same for all sub groups:
        for sub_group_name, sub_group in group.groups.items():
            NetCDF4._load_group(
                ds, path + sub_group_name, sub_group, fields
            )

    @expects_file_info(pos=2)
    def write(self, data, filename, **kwargs):
        """Save a xarray.Dataset to a NetCDF4 file

        Args:
            data: A xarray.Dataset object. It may contain 'pseudo' groups (i.e.
                variables with */* in their names). Those variables will be
                saved in subgroups.
            filename: A string or a :class:`FileInfo` object with the path
                where the data should be stored.

        Returns:
            None

        Examples:

            .. code-block:: python

                from typhon.files import NetCDF4
                import xarray as xr

                fh = NetCDF4()
                data = xr.Dataset({
                    "data/temperature": ("time", [0, 1]),
                    "lat": ("time", [0, 1]),
                    "lon": ("time", [0, 1]),
                }, coords={
                    "time": ("time", [0, 1]),
                })

                # Save the dataset:
                fh.write(data, "filename.nc")
        """
        group_vars = defaultdict(list)

        # Get the variables for the different groups:
        for full_name in data.variables:
            group, _ = NetCDF4._split_path(full_name)
            group_vars[group].append(full_name)

        # If we ware writing out multiple groups, we do not want to overwrite
        # the last file:
        user_mode = kwargs.pop("mode", "w")
        already_opened = False
        for group, variables in group_vars.items():
            ds = data[variables]

            # We do not want to store global coordinates in each subgroup.
            # Hence, we drop them before saving the group:
            coords_to_drop = [
                coord
                for coord in ds.coords.keys()
                if NetCDF4._split_path(coord)[0] != group
            ]
            ds = ds.drop(coords_to_drop)

            # Remove the group name from all variables (incl. dimensions):
            mapping = {
                full: NetCDF4._split_path(full)[1]
                for full in ds.data_vars
            }
            ds = ds.rename(mapping)
            mapping = {
                dim: NetCDF4._split_path(dim)[1]
                for dim in ds.dims
            }
            ds = ds.rename(mapping)

            ds.to_netcdf(
                filename.path, group=group,
                mode="a" if already_opened else user_mode,
                **kwargs
            )
            already_opened = True

    @staticmethod
    def _split_path(path):
        if "/" not in path:
            return None, path
        return path.rsplit("/", 1)


class Plotter(FileHandler):
    """File handler that can save matplotlib.figure objects to a file.

    This is a specialised file handler object that can just store
    matplotlib.figure objects. It cannot read from a file nor get the time
    coverage from one. This is designed for having a simple plot dataset as
    output.
    """

    def __init__(self, fig_args=None, **kwargs):
        """Initializes a Plotter file handler class.

        Args:
            fig_args: A dictionary of additional keyword arguments for the
                fig.savefig method.
        """
        # Call the base class initializer
        super().__init__(**kwargs)

        if fig_args is None:
            self.fig_args = {}
        else:
            self.fig_args = fig_args

    @expects_file_info(pos=2)
    def write(self, figure, filename, fig_args=None):
        """ Saves a matplotlib.figure object to a file.

        Args:
            figure: A matplotlib.figure object.
            filename: Path and name of the file as string or FileInfo object.
            fig_args: A dictionary of additional keyword arguments for the
                fig.savefig method. This updates the *fig_args* given during
                initialisation.
        """

        params = self.fig_args.copy()
        if fig_args is not None:
            params.update(**fig_args)

        return figure.savefig(filename.path, **params)


# class Numpy(handlers.FileHandler):
#     def __init__(self, **kwargs):
#         # Call the base class initializer
#         super().__init__(**kwargs)
#
#     def get_info(self, filename):
#         # Get info parameters from a file (time coverage, etc)
#         ...
#
#     def read(self, filename, fields=None):
#         """ Reads and parses files with numpy arrays and load them to a xarray.
#
#         See the base class for further documentation.
#         """
#         numpy_data = np.load(filename)
#         print(numpy_data.keys())
#         data = xarray.Dataset.from_dict(numpy_data)
#
#         return data
#
#     def write(self, filename, data):
#         """ Writes a xarray to a NetCDF file.
#
#         See the base class for further documentation.
#         """
#
#         # Data must be a xarray object!
#         data_dict = data.to_dict()
#         np.save(filename, data_dict)
#
#
# class Pickle(handlers.FileHandler):
#     def __init__(self, **kwargs):
#         # Call the base class initializer
#         super().__init__(**kwargs)
#
#     def get_info(self, filename):
#         # Get info parameters from a file (time coverage, etc)
#         ...
#
#     def read(self, filename, fields=None):
#         """ Reads and parses files with numpy arrays and load them to a xarray.
#
#         See the base class for further documentation.
#         """
#
#         with open(filename, 'rb') as file:
#             return pickle.load(file)
#
#     def write(self, filename, data):
#         """ Writes a xarray to a NetCDF file.
#
#         See the base class for further documentation.
#         """
#
#         with open(filename, 'wb') as file:
#             pickle.dump(data, file)
#
#
# class XML(handlers.FileHandler):
#     def __init__(self, **kwargs):
#         # Call the base class initializer
#         super().__init__(**kwargs)
#
#     def get_info(self, filename):
#         # Get info parameters from a file (time coverage, etc)
#         ...
#
#     def read(self, filename, fields=None):
#         """ Reads and parses NetCDF files and load them to a xarray.
#
#         See the parent class for further documentation.
#         """
#         #
#         return typhon.arts.xml.load(filename)
#
#     def write(self, filename, data):
#         """ Writes a xarray to a NetCDF file.
#
#         See the base class for further documentation.
#         """
#
#         # Data must be a xarray object!
#         typhon.arts.xml.save(data, filename)



# -*- coding: utf-8 -*-
"""Testing the functions in typhon.plots.colors.
"""
import filecmp
import os
from tempfile import mkstemp

import matplotlib.pyplot as plt
import matplotlib.colors as mcolors
import numpy as np
import pytest

from typhon.plots import colors


class TestColors:
    """Testing the cm functions."""
    ref_dir = os.path.join(os.path.dirname(__file__), "reference", "")

    def setup_method(self):
        """Create a temporary file."""
        fd, self.f = mkstemp()
        os.close(fd)

    def teardown_method(self):
        """Delete temporary file."""
        os.remove(self.f)

    def test_cmap2rgba(self):
        """Check colormap to RGB conversion."""
        ref = np.loadtxt(os.path.join(self.ref_dir, 'viridis.txt'),
                         comments='%')
        rgb = colors.cmap2rgba('viridis', 256)[:, :3]  # ignore alpha

        assert np.allclose(ref, rgb, atol=0.001)

    def test_cmap2rgba_interpolation(self):
        """Check colormap to RGBA interpolation."""
        max_planck_duplicates = np.array([
            [0., 0.4627451, 0.40784314, 1.],
            [0.48235294, 0.70980392, 0.67843137, 1.],
            [0.74901961, 0.85098039, 0.83137255, 1.],
            [0.96078431, 0.97254902, 0.97647059, 1.],
            [0.96078431, 0.97254902, 0.97647059, 1.],
        ])

        max_planck_interpolated = np.array([
            [0., 0.4627451, 0.40784314, 1.],
            [0.36318339, 0.64876586, 0.61158016, 1.],
            [0.6172549, 0.7812226, 0.75580161, 1.],
            [0.8038293, 0.88244521, 0.86892734, 1.],
            [0.96078431, 0.97254902, 0.97647059, 1.],
        ])

        assert np.allclose(
            max_planck_interpolated,
            colors.cmap2rgba('max_planck', 5, interpolate=True)
        )

        assert np.allclose(
            max_planck_duplicates,
            colors.cmap2rgba('max_planck', 5, interpolate=False)
        )

    def test_cmap2cpt(self):
        """Export colormap to cpt file."""
        colors.cmap2cpt('viridis', filename=self.f)
        ref = os.path.join(self.ref_dir, 'viridis.cpt')

        with open(self.f) as testfile, open(ref) as reffile:
            assert testfile.readlines() == reffile.readlines()

    def test_cmap2txt(self):
        """Export colormap to txt file."""
        colors.cmap2txt('viridis', filename=self.f)
        ref = os.path.join(self.ref_dir, 'viridis.txt')

        with open(self.f) as testfile, open(ref) as reffile:
            assert testfile.readlines() == reffile.readlines()

    def test_cmap2act(self):
        """Export colormap to act file."""
        colors.cmap2act('viridis', filename=self.f)
        ref = os.path.join(self.ref_dir, 'viridis.act')

        assert filecmp.cmp(self.f, ref)

    def test_cmap_from_txt(self):
        """Import colormap from txt file."""
        viridis = plt.get_cmap('viridis')
        cmap = colors.cmap_from_txt(os.path.join(self.ref_dir, 'viridis.txt'))

        plt.register_cmap(cmap=viridis)  # Register original viridis.

        idx = np.linspace(0, 1, 256)
        assert np.allclose(viridis(idx), cmap(idx), atol=0.001)

    def test_cmap_from_act(self):
        """Import colormap from act file."""
        viridis = plt.get_cmap('viridis')
        cmap = colors.cmap_from_act(os.path.join(self.ref_dir, 'viridis.act'))

        plt.register_cmap(cmap=viridis)  # Register original viridis.

        idx = np.linspace(0, 1, 256)
        assert np.allclose(viridis(idx), cmap(idx), atol=0.004)

    def test_get_material_design(self):
        """Test the retrieval of material design colors."""
        hex_color = colors.get_material_design('red', shade='500')
        assert hex_color == '#F44336'

        hex_colors = colors.get_material_design('red', shade=None)
        assert hex_colors == ['#FFEBEE', '#FFCDD2', '#EF9A9A', '#E57373',
                              '#EF5350', '#F44336', '#E53935', '#D32F2F',
                              '#C62828', '#B71C1C', '#FF8A80', '#FF5252',
                              '#FF1744', '#D50000']

    def test_get_material_design_valuerror(self):
        """Test the behavior for undefined material design colors or shades."""
        with pytest.raises(ValueError):
            colors.get_material_design('undefined_color')

        with pytest.raises(ValueError):
            colors.get_material_design('red', 'undefined_shade')

    def test_named_color_mapping(self):
        """Test if the typhon colors are available in the name mapping."""
        assert all([c in mcolors.get_named_colors_mapping()
                   for c in colors.TYPHON_COLORS.keys()])

    def test_named_color_hex(self):
        """Test if the 'ty:uhh-red' hex-value is correct."""
        assert mcolors.get_named_colors_mapping()['ty:uhh-red'] == '#ee1d23'



# -*- coding: utf-8 -*-

"""This module provides functions related to plot or to plot data.
"""

from typhon.plots import cm  # noqa
from typhon.plots.colors import *  # noqa
from typhon.plots.common import *  # noqa
from typhon.plots.formatter import *  # noqa
from typhon.plots.plots import *  # noqa
from typhon.plots.arts_lookup import *  # noqa
from typhon.plots.ppath import *  # noqa
try:
    from typhon.plots.maps import *  # noqa
except ImportError:
    pass

__all__ = [s for s in dir() if not s.startswith('_')]



r"""
This submodule contains implementations of different retrieval
methods as well as functions for the assessment of their performance.
"""

from .common import *  # noqa
from .oem import *  # noqa
from .spareice import *  # noqa



# -*- coding: utf-8 -*-

try:
    __ATRASU_SETUP__
except:
    __ATRASU_SETUP__ = False

if not __ATRASU_SETUP__:
	from . import spectra
	from . import setup_atmosphere
	from . import const
	from . import nonltecalc
	from . import mathmatics
	from . import rtc



# -*- coding: utf-8 -*-
"""Testing the functions in typhon.utils.
"""
import warnings
import numpy
import xarray
from time import sleep
from datetime import timedelta

import pytest

from typhon import utils


class TestUtils:
    """Testing the typhon.utils functions."""
    def test_deprecated(self):
        """Test deprecation warning."""
        @utils.deprecated
        def func():
            pass

        with warnings.catch_warnings():
            warnings.simplefilter('error')
            with pytest.raises(DeprecationWarning):
                func()

    def test_image2mpeg(self):
        """Test the behavior when no files are found."""
        with pytest.raises(Exception):
            utils.image2mpeg(glob='', outfile='foo.mp4')

    def test_unique(self):
        assert utils.unique([0, 5, 1, 2, 0, 3, 1]) == [0, 5, 1, 2, 3]

    def test_undo_xarray_floatification(self):
        ds = xarray.Dataset(
            {"a": (["x"], numpy.array([1, 2, 3], dtype="f4")),
             "b": (["x"], numpy.array([2.0, 3.0, 4.0])),
             "c": (["x"], numpy.array(["2010-01-01", "2010-01-02",
                                       "2010-01-03"], dtype="M8"))})
        ds["a"].encoding = {"dtype": numpy.dtype("i4"),
                            "_FillValue": 1234}
        # c should NOT be converted because it's a time
        ds["c"].encoding = {"dtype": numpy.dtype("i8"),
                            "_FillValue": 12345}
        ds2 = utils.undo_xarray_floatification(ds)
        assert ds is not ds2  # has to be a copy
        assert ds["a"].encoding == ds2["a"].encoding
        assert numpy.allclose(ds["a"], ds2["a"])
        assert ds2["a"].dtype == ds2["a"].encoding["dtype"]
        assert (ds2["c"] == ds["c"]).all()
        assert ds2["c"].dtype == ds["c"].dtype
        assert ds2["b"].dtype == ds["b"].dtype

    def test_Timer_methods(self):
        """Test initialization and basic methods of the `Timer` class."""
        t = utils.Timer(verbose=False)
        t.start()
        sleep(0.05)
        dt = t.stop()

        assert timedelta(seconds=0.05) <= dt

    def test_Timer_block(self):
        """Test usage of a `Timer` object in a with-block."""
        with utils.Timer(verbose=False):
            ...

    def test_Timer_decorator(self):
        """Test usage of a `Timer` object as function decorator."""
        @utils.Timer(verbose=False)
        def foo():
            return

        foo()



import numpy as np
import scipy as sp
import matplotlib.pyplot as plt
from typhon.arts.catalogues import Sparse
import ctypes as c

class Block(object):
    """
    A block of a covariance matrix.

    A covariance matrix block holds the covariances of a given retrieval
    quantity or the covariances between todifferent retrieval quantities. A
    block is identified by its row and column block indices. The row and column
    indices of the block correspond to the retrieval quantity indices within
    ARTS. A covariance matrix block consists of a matrix holding the covariances,
    which may be either dense or sparse, as well as additional information on
    the location of the block in the full covariance matrix.
    """

    @classmethod
    def __from_covariance_matrix_block_struct__(cls, s, inverse):
        """
        Create a block from a :class:`CovarianceMatrixBlockStruct`
        returned from the ARTS API.

        Paramters:
            s: The :class:`CovarianceMatrixBlockStruct` to create the
               block from.
            inverse: Flag that indicates whether the block belongs to
               the normal part of the covariance matrix or its inverse.

        Returns: The :class:`Block` object that represents the given
            :class:`CovarianceMatrixBlockStruct`
        """

        i, j   = list(s.indices)
        rs, cs = list(s.position)
        m, n   = list(s.dimensions)

        if not s.inner_ptr:
            matrix = np.ctypeslib.as_array(c.cast(s.ptr, c.POINTER(c.c_double)), (m, n))
        else:
            nnz = s.nnz
            data = np.ctypeslib.as_array(c.cast(s.ptr, c.POINTER(c.c_double)),
                                         (nnz,))
            row_indices = np.ctypeslib.as_array(s.inner_ptr, (nnz,))
            col_starts  = np.ctypeslib.as_array(s.outer_ptr, (m + 1,))
            matrix = sp.sparse.csr_matrix((data, row_indices, col_starts),
                                          shape = (m, n))
        return Block(i, j, rs, cs, inverse, matrix)

    def __init__(self, i, j, row_start, column_start, inverse, matrix):
        """
        Parameters:
            i(int): The row-block index of the covariance matrix block.
            j(int): The column-block index of the covariance matrix block.
            row_start(int): Row index of the left- and uppermost element in
                in the block.
            column_start(int): Column index of the left- and uppermost element
                in the block
            inverse(bool): Flag indicating whether the block is part of the
                inverse of the covariance matrix or not.
            matrix(np.ndarray or sp.sparse): The matrix of which the block
                consists.

        """
        self._i = i
        self._j = j
        self._row_start    = row_start
        self._column_start = column_start
        self._inverse = inverse
        self._matrix = matrix

    #
    # Read-only properties
    #

    @property
    def i(self):
        """Row-index of the block."""
        return self._i

    @property
    def j(self):
        """Column-index of the block"""
        return self._j

    @property
    def row_start(self):
        """Row-index of the left- and uppermost element of the
            block w.r.t to the full covariance matrix."""
        return self._row_start

    @property
    def column_start(self):
        """Column index of the left- and uppermost element of
            the block w.r.t. to the full covariance matrix."""
        return self._column_start

    @property
    def inverse(self):
        """Flag that indicates whether this block is part of
            the normal part of the covariance matrix or its inverse."""
        return self._inverse

    @property
    def matrix(self):
        """The matrix containing the covariances of the block."""
        return self._matrix

    def write_xml(self, xmlwriter, attr = None):
        """
        Serialize block and write to xml stream.

        Opens a new tag for the block and writes the matrix into
        it. Attributes of the block are saved as attributes of
        the newly create block.

        Parameters:
            xmlwriter: The xml stream to which to write the block.
            attr(dict): Additional attributes that should be added
                the tag that is created for the block.
        """
        if attr is None:
            attr = {}

        attr["row_index"] = self.i
        attr["column_index"] = self.j
        attr["row_start"] = self.row_start
        attr["column_start"] = self.column_start
        attr["row_extent"], attr["column_extent"] = self.matrix.shape
        attr["is_inverse"] = int(self.inverse)

        if sp.sparse.issparse(self.matrix):
            if not type(self.matrix) == Sparse:
                # why? because I can ...
                self.matrix.__class__.write_xml = Sparse.write_xml
            attr["type"] = "Sparse"
        else:
            attr["type"] = "Dense"

        xmlwriter.open_tag('Block', attr)
        xmlwriter.write_xml(self.matrix)
        xmlwriter.close_tag()

class CovarianceMatrix(object):
    """
    Covariance matrix class representing the ARTS group of the same name
    implementing covariance matrices for OEM calculations in ARTS. Covariance
    matrices are stored as block diagonal matrices where each block represents
    covariances between two retrieval quantities.

    Since covariance matrices must be symmetric only blocks lying on or above
    the diagonal are stored. The covariance matrix class is designed to hold
    both, the covariance matrix and its inverse. This has the advantage that
    the inverse of the covariance matrix can be set directly by the user, which
    is useful for Tikhonov regularization and when the inverse is available in
    closed form.
    """
    #
    # Class methods
    #

    @classmethod
    def __from_variable_value_struct__(cls, s):
        """
        Implements ARTS-API interface for returning objects from
        an ARTS workspace.


        """
        from typhon.arts.workspace.api import arts_api

        n_blocks = s.dimensions[0]
        n_inv_blocks = s.dimensions[1]

        blocks = []
        for i in range(n_blocks):
            bs = arts_api.get_covariance_matrix_block(s.ptr, i, False)
            b = Block.__from_covariance_matrix_block_struct__(bs, False)
            blocks += [b]

        inv_blocks = []
        for i in range(n_inv_blocks):
            bs = arts_api.get_covariance_matrix_block(s.ptr, i, True)
            b = Block.__from_covariance_matrix_block_struct__(bs, True)
            inv_blocks += [b]

        return CovarianceMatrix(blocks, inv_blocks)


    @classmethod
    def from_xml(cls, xmlelement):
        """Load a covariance matrix from an ARTS XML fiile.

        Returns:
           The loaded covariance matrix as :class:`CovarianceMatrix` object
        """
        n_blocks = xmlelement.get("n_blocks")

        blocks = []
        inv_blocks = []
        for b in list(xmlelement):

            i = b.get("row_index")
            j = b.get("column_index")
            row_start    = int(b.get("row_start"))
            column_start = int(b.get("column_start"))
            inverse = bool(int(b.get("is_inverse")))
            matrix = b[0].value()

            b = Block(i, j, row_start, column_start, inverse, matrix)
            if inverse:
                inv_blocks += [b]
            else:
                blocks += [b]
        return CovarianceMatrix(blocks, inv_blocks)

    def __init__(self, blocks, inverse_blocks = [], workspace = None):
        """
        Create a covariance matrix object.

        Parameters:
            blocks(list): List containing the blocks that make up the
                covariance matrix.
            inverse_blocks: Blocks that make up the inverse of the
                covariance. Can be provided to avoid computation of
                the inverse of the covariance matrix.
            workspace: :class:`Workspace` to associate the covariance
                matrix to.
        """
        self._blocks         = blocks
        self._inverse_blocks = inverse_blocks
        self._workspace      = None

    #
    # Read-only properties
    #

    @property
    def blocks(self):
        """ The blocks contained in the covariance matrix."""
        return self._blocks

    @property
    def inverse_blocks(self):
        """The blocks that contained in the inverse of the covariance matrix."""
        return self._inverse_blocks

    @property
    def workspace(self):
        """The workspace associated with the covariance matrix."""
        return self._workspace

    #
    # Serialization
    #

    def write_xml(self, xmlwriter, attr = None):
        """
        Implements typhon xml serialization interface.

        Parameters:
            xmlwriter: The xml stream to which to write the block.
            attr(dict): Additional attributes that should be added
                the tag that is created for the block.
        """

        if attr is None:
            attr = {}

        attr['n_blocks'] = len(self.blocks) + len(self.inverse_blocks)
        xmlwriter.open_tag('CovarianceMatrix', attr)

        for b in self.blocks:
            xmlwriter.write_xml(b)
        for b in self.inverse_blocks:
            xmlwriter.write_xml(b)

        xmlwriter.close_tag()

    def to_dense(self):
        """Conversion to dense representation.

        Converts the covariance matrix to a 2-dimensional numpy.ndarray.

        Returns:
            The covariance matrix as dense matrix.

        """
        m = max([b.row_start + b.matrix.shape[0] for b in self.blocks])
        n = max([b.column_start + b.matrix.shape[1] for b in self.blocks])
        mat = np.zeros((m, n))
        for b in self.blocks:
            m0 = b.row_start
            n0 = b.column_start
            dm = b.matrix.shape[0]
            dn = b.matrix.shape[1]
            if sp.sparse.issparse(b.matrix):
                mat[m0 : m0 + dm, n0 : n0 + dn] = b.matrix.toarray()
            else:
                mat[m0 : m0 + dm, n0 : n0 + dn] = b.matrix
        return mat

def plot_covariance_matrix(covariance_matrix, ax = None):
    """
    Plots a covariance matrix.

    Parameters:
        covariance_matrix(:class:`CovarianceMatrix`): The covariance matrix
            to plot
        ax(matplotlib.axes): An axes object into which to plot the
            covariance matrix.
    """

    if ax is None:
        ax = plt.gca()

    for b in covariance_matrix.blocks:
        y = np.arange(b.row_start, b.row_start + b.matrix.shape[0] + 1) - 0.5
        x = np.arange(b.column_start, b.column_start + b.matrix.shape[1] + 1) - 0.5
        ax.pcolormesh(x, y, np.array(b.matrix.toarray()))



    m = max([b.row_start + b.matrix.shape[0] for b in covariance_matrix.blocks])
    n = max([b.column_start + b.matrix.shape[1] for b in covariance_matrix.blocks])
    ax.set_xlim([-0.5, n + 0.5])
    ax.set_ylim([m + 0.5, -0.5])
